# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
from typing import Any

import jax
import jax.experimental.pallas.tpu as pltpu
import jax.numpy as jnp
from jax import core
from jax.experimental import mosaic
from jax.experimental.mosaic.dialects import tpu
from jaxlib.mlir import ir
from jaxlib.mlir.dialects import arith, func, memref, scf, vector


class VectorTypeHelper:
    """Helper to create VectorType with a specific element type."""

    def __init__(self, element_type_fn):
        self.element_type_fn = element_type_fn

    def __getitem__(self, shape):
        if isinstance(shape, int):
            shape = [shape]
        return ir.VectorType.get(shape, self.element_type_fn())


_I32 = VectorTypeHelper(lambda: ir.IntegerType.get_signless(32))
_F32 = VectorTypeHelper(lambda: ir.F32Type.get())
_BF16 = VectorTypeHelper(lambda: ir.BF16Type.get())


def is_supported_by_sc_gather_reduce(x_shape: int,
                                     sc_kernel_threshold: int) -> bool:
    if x_shape > sc_kernel_threshold and pltpu.get_tpu_info().generation == 7:
        return True
    return False


@jax.jit(static_argnames=[
    "reduce_group_size",
    "single_sc",
    "col_chunk_size",
    "loop_unroll_factor_1",
    "loop_unroll_factor_2",
    "loop_unroll_factor_3",
    "loop_parallel_access_1",
    "loop_parallel_access_2",
    "loop_parallel_access_3",
], )
def sc_gather_reduce(
    op: jax.Array,
    idx: jax.Array,
    topk_weights: jax.Array | None = None,
    *,
    reduce_group_size: int,
    single_sc: bool = False,
    col_chunk_size: int = int(3.5 * 1024),  # half a row
    row_chunk_size: int = 16,  # writing back 2 rows given reduce size of 8
    loop_unroll_factor_1: int = 8,
    loop_unroll_factor_2: int = 2,
    loop_unroll_factor_3: int = 8,
    loop_parallel_access_1: bool = True,
    loop_parallel_access_2: bool = True,
    loop_parallel_access_3: bool = False,
) -> jax.Array:
    """Performs a gather-reduce operation on SparseCore.

  This kernel groups rows of the operand `op` based on `idx`, sums them up,
  and scatters the results. The gather and add operations are performed in fp32,
  and the results are written back in bf16.

  Args:
    op: The operand matrix in fp32 [B, K] to reduce.
    idx: The indices in int32[M,] guiding the reduction and scatter.
    topk_weights: Optional weights to apply to the gathered operands.
    reduce_group_size: The size of the groups to reduce.
    single_sc: Whether to use a single SparseCore.
    col_chunk_size: The size of column chunks to process.
    row_chunk_size: The size of row chunks for internal processing.

  Returns:
    The result of operation, bf16 matrix [M/reduce_group_size x K].
  """

    assert (op.dtype == jnp.float32 or op.dtype == jnp.bfloat16
            ), f"op.dtype must be f32 or bf16, but got {op.dtype}"
    is_bf16 = op.dtype == jnp.bfloat16
    assert op.shape[0] % reduce_group_size == 0, (
        "op.shape[0] must be divisible by reduce_group_size, but got"
        f" {op.shape[0]} and {reduce_group_size}")
    assert (  # writing back in bf16 (2 rows at once)
        row_chunk_size / reduce_group_size == 2
    ), (f"row_chunk_size must be 2 * reduce_group_size, but got {row_chunk_size=}"
        f" and {reduce_group_size=}")

    if topk_weights is not None:
        assert topk_weights.ndim == 2
        assert topk_weights.shape[0] * 128 == idx.shape[0]
        assert topk_weights.shape[1] == 128

    tpu_info = pltpu.get_tpu_info()
    if tpu_info.sparse_core is None:
        raise ValueError("SparseCore is not available on this TPU version.")

    used_sc_cores = 1 if single_sc else 2
    num_sc_per_core = tpu_info.sparse_core.num_subcores
    num_sc = num_sc_per_core * used_sc_cores

    vreg_size = tpu_info.sparse_core.num_lanes

    from jax.interpreters import mlir
    with mlir.make_ir_context() as ctx, ir.Location.unknown():
        # The TPU dialect is required for its TPU_DimensionSemantics
        tpu.register_dialect(ctx)

        bf16 = ir.BF16Type.get()
        i32 = ir.IntegerType.get_signless(32)
        f32 = ir.F32Type.get()
        index = ir.IndexType.get()

        memoryspace_tilespmem = ir.Attribute.parse("#tpu.memory_space<vmem>")
        memoryspace_hbm = ir.Attribute.parse("#tpu.memory_space<hbm>")

        memory_space_semaphore = ir.Attribute.parse(
            "#tpu.memory_space<semaphore_mem>")
        dma_semaphore_type = ir.Type.parse("!tpu.dma_semaphore")

        # A mask of all True values to enable all sublanes in vector operations.
        enable_all_sublanes_mask = ir.DenseBoolArrayAttr.get([True] *
                                                             vreg_size)

        assert row_chunk_size == vreg_size

        def _kernel_impl(
            current_sc_core,
            current_local_core,
            idx_ref,
            op_ref,
            weights_ref,
            out_ref,
            func_op,
        ):
            constants: dict[tuple[Any, Any], Any] = {}

            def const_lut(val, ty=None):
                ty = index if ty is None else ty
                if (val, ty) not in constants:
                    with ir.InsertionPoint.at_block_begin(func_op.entry_block):
                        constants[(val, ty)] = arith.constant(
                            ty, ir.IntegerAttr.get(ty, val))
                return constants[(val, ty)]

            def fill_load_offset_tile(offset_tile_local, idx_tile_local,
                                      col_pos):
                """Fills the offset tile for indirect DMA gather.

        This function calculates the HBM offsets from which to gather rows
        based on the indices in idx_tile_local, for a given column chunk.
        The offsets are calculated to correctly index into the operand `op`
        in HBM, considering the memory layout and the current column chunk
        being processed. The calculated offsets are stored in
        offset_tile_local, which is later used by tpu.enqueue_indirect_dma.

        Args:
          offset_tile_local: The destination memref in TileSpMem to store
            calculated offsets.
          idx_tile_local: A memref in TileSpMem containing a chunk of indices
            of rows to gather from `op`.
          col_pos: The index of the current column chunk being processed.

        Returns:
          The offset_tile_local memref filled with offsets for DMA gather.
        """
                idx_loaded = tpu.load(
                    _I32[row_chunk_size],
                    idx_tile_local,
                    [const_lut(0)],
                    enable_all_sublanes_mask,
                )
                parity = arith.remui(
                    idx_loaded,
                    vector.broadcast(_I32[row_chunk_size], const_lut(2, i32)))
                if is_bf16:
                    mul_mem_layout = 4
                else:
                    mul_mem_layout = 8

                iota = arith.constant(
                    _I32[row_chunk_size],
                    ir.DenseIntElementsAttr.get(
                        array.array("i", [
                            i * mul_mem_layout for i in range(row_chunk_size)
                        ]),
                        type=_I32[row_chunk_size],
                    ),
                )
                loop_over_idx = scf.ForOp(
                    lower_bound=const_lut(0),
                    upper_bound=const_lut(row_chunk_size + 1),
                    step=const_lut(1),
                )
                loop_over_idx.attributes[
                    "sc.loop_unroll_factor"] = ir.IntegerAttr.get(
                        i32, loop_unroll_factor_2)
                if loop_parallel_access_2:
                    loop_over_idx.attributes[
                        "sc.parallel_access"] = ir.UnitAttr.get()
                with ir.InsertionPoint(loop_over_idx.body):
                    i = loop_over_idx.induction_variable
                    base_idx = vector.extract(
                        idx_loaded,
                        dynamic_position=[i],
                        static_position=ir.DenseI64ArrayAttr.get(
                            [-9223372036854775808]),
                    )

                    # align base_idx with memory layout
                    base_idx_div = arith.divui(base_idx, const_lut(8, i32))
                    if is_bf16:
                        base_idx_rem = arith.divui(
                            arith.remui(base_idx, const_lut(8, i32)),
                            const_lut(2, i32))
                    else:
                        base_idx_rem = arith.remui(base_idx, const_lut(8, i32))
                    base_idx = arith.addi(
                        arith.muli(
                            base_idx_div,
                            const_lut(op.shape[1] // 128 * mul_mem_layout,
                                      i32),
                        ),  # go/NOTYPO
                        base_idx_rem,
                    )

                    # consider col chunk
                    base_idx = arith.addi(
                        base_idx,
                        arith.muli(  # go/NOTYPO
                            arith.index_cast(i32, col_pos),
                            const_lut(col_chunk_size // 128 * mul_mem_layout,
                                      i32),
                        ),
                    )

                    start_vec = arith.addi(
                        vector.broadcast(_I32[row_chunk_size], base_idx), iota)
                    loop_j = scf.ForOp(
                        const_lut(0),
                        const_lut((col_chunk_size // 128) // vreg_size),
                        const_lut(1),
                        iter_args=[start_vec],
                    )
                    loop_j.attributes[
                        "sc.loop_unroll_factor"] = ir.IntegerAttr.get(
                            i32, loop_unroll_factor_3)
                    if loop_parallel_access_3:
                        loop_j.attributes[
                            "sc.parallel_access"] = ir.UnitAttr.get()
                    with ir.InsertionPoint(loop_j.body):
                        vec = loop_j.inner_iter_args[0]

                        idx_local = arith.addi(
                            arith.muli(i, const_lut(
                                (col_chunk_size // 128))),  # go/NOTYPO
                            arith.muli(  # go/NOTYPO
                                loop_j.induction_variable,
                                const_lut(vreg_size)),
                        )
                        tpu.store(
                            vec,
                            offset_tile_local,
                            [idx_local],
                            enable_all_sublanes_mask,
                        )
                        next_vec = arith.addi(
                            vec,
                            vector.broadcast(
                                _I32[row_chunk_size],
                                const_lut(mul_mem_layout * vreg_size, i32),
                            ),
                        )
                        scf.YieldOp([next_vec])

                    rem_vreg = col_chunk_size / 128 % vreg_size / vreg_size

                    last_full_vec = tpu.load(
                        _I32[vreg_size],
                        offset_tile_local,
                        [
                            arith.subi(
                                arith.muli(const_lut(
                                    (col_chunk_size // 128)), i),  # go/NOTYPO
                                const_lut(int(vreg_size * (1 + rem_vreg))),
                            )
                        ],
                        enable_all_sublanes_mask,
                    )
                    add_vec = vector.broadcast(
                        _I32[vreg_size],
                        const_lut(int(mul_mem_layout * rem_vreg * vreg_size),
                                  i32),
                    )  # 4 comes from bf16

                    last_full_vec = arith.addi(last_full_vec, add_vec)
                    tpu.store(
                        last_full_vec,
                        offset_tile_local,
                        [
                            arith.subi(
                                arith.muli(const_lut(
                                    (col_chunk_size // 128)), i),  # go/NOTYPO
                                const_lut(int(vreg_size)),
                            )
                        ],
                        enable_all_sublanes_mask,
                    )

                    scf.YieldOp([])
                return offset_tile_local, parity

            def load_weights(lin_idx, dst_tile, sflag):
                """Loads weights from HBM to TileSpMem.

        This function calculates offsets for reading weights from HBM based on
        a linear index `lin_idx`. It handles a specific packed memory layout
        of weights where weights for two rows are packed together (bf16). It
        calculates parity to determine which part of the packed data to use
        later in `perform_add`. It uses indirect DMA to load weights into
        dst_tile.

        Args:
          lin_idx: Linear index used to calculate weight offsets in HBM.
          dst_tile: The destination memref in TileSpMem to load weights into.
          sflag: The semaphore to use for the DMA operation.

        Returns:
          Parity bit (0 or 1) indicating which set of weights to use from
          the loaded packed data.
        """
                lin_idx = arith.divui(
                    lin_idx, const_lut(16))  # reading in increments of 16

                # parity
                # (i >> 3) & 1
                parity = arith.andi(arith.shrui(lin_idx, const_lut(3)),
                                    const_lut(1))

                # address
                # (i & 7) | ((i >> 4) << 3)
                lin_idx = arith.ori(
                    arith.andi(lin_idx, const_lut(7)),
                    arith.shli(arith.shrui(lin_idx, const_lut(4)),
                               const_lut(3)),
                )

                # lin_idx is index type, cast to i32
                lin_idx_i32 = arith.index_cast(i32, lin_idx)
                lin_idx_base = arith.muli(lin_idx_i32, const_lut(16, i32))
                lin_idx_base_vec = vector.broadcast(_I32[row_chunk_size],
                                                    lin_idx_base)

                iota = arith.constant(
                    _I32[row_chunk_size],
                    ir.DenseIntElementsAttr.get(
                        array.array("i", [i for i in range(16)]),
                        type=_I32[row_chunk_size],
                    ),
                )
                offsets = arith.addi(lin_idx_base_vec, iota)

                tpu.store(
                    offsets,
                    offset_tile_weights,
                    [const_lut(0)],
                    enable_all_sublanes_mask,
                )

                tpu.enqueue_indirect_dma(
                    source=weights_ref,
                    target=dst_tile,
                    offsets=offset_tile_weights,
                    semaphore=sflag,
                )
                tpu.wait_indirect_dma(semaphore=sflag,
                                      src=weights_ref,
                                      dst=dst_tile)

                return parity

            def perform_add(
                scratch_local,
                scratch_out_local,
                idx_parity,
                weights_local=None,
                parity=None,
            ):
                """Performs reduction (summation) of rows in scratchpad.

        This function reduces `row_chunk_size` rows stored in scratch_local
        into 2 rows by summing rows in groups of `reduce_group_size`. If
        weights_local is provided, rows are multiplied by weights before
        summation. The result is packed into bf16 format and stored in
        scratch_out_local. With row_chunk_size=16 and reduce_group_size=8,
        this reduces 16 rows to 2 rows (rows 0-7 summed to one row, rows 8-15
        summed to another).

        Args:
          scratch_local: Memref in TileSpMem containing rows gathered from `op`.
          scratch_out_local: Memref in TileSpMem to store the 2 reduced rows
            in bf16 format.
          weights_local: Optional memref in TileSpMem containing weights to
            apply before reduction.
          parity: Optional parity bit indicating which weights to use from
            packed weight data in weights_local.

        Returns:
          The scratch_out_local memref containing reduced rows.
        """
                weights_vecs = None
                if weights_local is not None:
                    f32_zero = arith.constant(f32, ir.FloatAttr.get(f32, 0.0))
                    zero_vec_f32 = vector.broadcast(_F32[vreg_size], f32_zero)
                    # Load 32 weights (64B) -> 16 weights for R0 + 16 weights for R1 from
                    # packed layout. We check parity to decide which one to take. We need
                    # to extract the even or odd elements
                    weights_local_2x16 = memref.reinterpret_cast(
                        ir.MemRefType.get(
                            (2, row_chunk_size),
                            bf16,
                            ir.StridedLayoutAttr.get(0, [row_chunk_size, 1]),
                            memory_space=memoryspace_tilespmem,
                        ),
                        weights_local,
                        [],
                        [],
                        [],
                        static_offsets=[0],
                        static_sizes=[2, row_chunk_size],
                        static_strides=[row_chunk_size, 1],
                    )
                    raw_weights = tpu.load(
                        _BF16[2, row_chunk_size],
                        weights_local_2x16,
                        [const_lut(0), const_lut(0)],
                        enable_all_sublanes_mask,
                    )

                    # Reminder Mosaic SC compiler: 'compressed' flips [d0, d1] to [d1, d0]
                    raw_weights_16x2 = vector.shape_cast(
                        _BF16[16, 2], raw_weights)
                    # part=1 corresponds to sc_tpu.unpackf result 1 (Even indices)
                    weights_evens = tpu.unpack_subelements(
                        _F32[16],
                        raw_weights_16x2,
                        0,
                        ir.Attribute.parse("#tpu.pack_format<compressed>"),
                    )
                    # part=0 corresponds to sc_tpu.unpackf result 0 (Odd indices)
                    weights_odds = tpu.unpack_subelements(
                        _F32[16],
                        raw_weights_16x2,
                        1,
                        ir.Attribute.parse("#tpu.pack_format<compressed>"),
                    )

                    is_odd = arith.cmpi(
                        arith.CmpIPredicate.eq,
                        parity,
                        const_lut(0),
                    )
                    raw_weights_f32 = arith.select(is_odd, weights_odds,
                                                   weights_evens)

                    weights_vecs = [
                        vector.broadcast(
                            _F32[vreg_size],
                            vector.extract(
                                raw_weights_f32,
                                dynamic_position=[],
                                static_position=ir.DenseI64ArrayAttr.get([i]),
                            ),
                        ) for i in range(16)
                    ]

                # reinterpret cast to correct shape to read from it
                if not is_bf16:
                    new_scratch_shape = (row_chunk_size, col_chunk_size)
                    new_scratch_layout = ir.StridedLayoutAttr.get(
                        0, [col_chunk_size, 1])
                    new_scratch_ref_ty = ir.MemRefType.get(
                        new_scratch_shape,
                        f32,
                        new_scratch_layout,
                        memory_space=memoryspace_tilespmem,
                    )
                    scratch_local = memref.reinterpret_cast(
                        new_scratch_ref_ty,
                        scratch_local,
                        [],
                        [],
                        [],
                        static_offsets=[0],
                        static_sizes=new_scratch_shape,
                        static_strides=[col_chunk_size, 1],
                    )

                loop_pack_scratch = scf.ForOp(const_lut(0),
                                              const_lut(col_chunk_size),
                                              const_lut(vreg_size))
                loop_pack_scratch.attributes[
                    "sc.loop_unroll_factor"] = ir.IntegerAttr.get(
                        i32, loop_unroll_factor_1)
                if loop_parallel_access_1:
                    loop_pack_scratch.attributes[
                        "sc.parallel_access"] = ir.UnitAttr.get()
                with ir.InsertionPoint(loop_pack_scratch.body):
                    col_offset = loop_pack_scratch.induction_variable

                    # map col_offset to idx and col inside 128
                    if is_bf16:
                        row_add = arith.divui(col_offset, const_lut(128))
                        col_pos = arith.remui(col_offset, const_lut(128))
                    else:
                        row_add = None
                        col_pos = None

                    def get_row_val(row_idx):
                        if is_bf16:
                            row_idx_l = row_idx * (col_chunk_size // 128)
                            vec_bf16_2x16 = tpu.load(
                                _BF16[2, 16],
                                scratch_local,
                                [
                                    arith.addi(const_lut(row_idx_l), row_add),
                                    arith.muli(col_pos, const_lut(2)),
                                ],
                                enable_all_sublanes_mask,
                            )
                            vec_bf16_16x2 = vector.shape_cast(
                                _BF16[16, 2], vec_bf16_2x16)
                            vec_f32_evens = tpu.unpack_subelements(
                                _F32[16],
                                vec_bf16_16x2,
                                0,
                                ir.Attribute.parse(
                                    "#tpu.pack_format<compressed>"),
                            )
                            vec_f32_odds = tpu.unpack_subelements(
                                _F32[16],
                                vec_bf16_16x2,
                                1,
                                ir.Attribute.parse(
                                    "#tpu.pack_format<compressed>"),
                            )
                            parity_of_row = vector.extract(
                                idx_parity,
                                dynamic_position=[],
                                static_position=ir.DenseI64ArrayAttr.get(
                                    [row_idx]),
                            )
                            is_odd_scalar = arith.cmpi(arith.CmpIPredicate.eq,
                                                       parity_of_row,
                                                       const_lut(0, i32))
                            return arith.select(is_odd_scalar, vec_f32_odds,
                                                vec_f32_evens)
                        else:
                            return tpu.load(
                                _F32[vreg_size],
                                scratch_local,
                                [const_lut(row_idx), col_offset],
                                enable_all_sublanes_mask,
                            )

                    row0 = get_row_val(0)
                    if weights_local is not None:
                        row0 = arith.mulf(row0, weights_vecs[0])
                        row0 = arith.select(
                            arith.cmpf(arith.CmpFPredicate.OEQ,
                                       weights_vecs[0], zero_vec_f32),
                            zero_vec_f32, row0)

                    row8 = get_row_val(8)
                    if weights_local is not None:
                        row8 = arith.mulf(row8, weights_vecs[8])
                        row8 = arith.select(
                            arith.cmpf(arith.CmpFPredicate.OEQ,
                                       weights_vecs[8], zero_vec_f32),
                            zero_vec_f32, row8)

                    for sum_idx in range(7):
                        tmp_row0 = get_row_val(sum_idx + 1)
                        if weights_local is not None:
                            tmp_row0 = arith.mulf(tmp_row0,
                                                  weights_vecs[sum_idx + 1])
                            tmp_row0 = arith.select(
                                arith.cmpf(arith.CmpFPredicate.OEQ,
                                           weights_vecs[sum_idx + 1],
                                           zero_vec_f32), zero_vec_f32,
                                tmp_row0)

                        row0 = arith.addf(row0, tmp_row0)

                        tmp_row8 = get_row_val(8 + sum_idx + 1)
                        if weights_local is not None:
                            tmp_row8 = arith.mulf(
                                tmp_row8, weights_vecs[8 + sum_idx + 1])
                            tmp_row8 = arith.select(
                                arith.cmpf(arith.CmpFPredicate.OEQ,
                                           weights_vecs[8 + sum_idx + 1],
                                           zero_vec_f32), zero_vec_f32,
                                tmp_row8)

                        row8 = arith.addf(row8, tmp_row8)

                    packed_16x2 = tpu.pack_subelements(
                        _BF16[vreg_size, 2],
                        [row8, row0],
                        [0, 1],
                        ir.Attribute.parse("#tpu.pack_format<compressed>"),
                    )
                    packed = vector.shape_cast(_BF16[2, vreg_size],
                                               packed_16x2)

                    tpu.store(
                        packed,
                        scratch_out_local,
                        [const_lut(0),
                         arith.muli(col_offset, const_lut(2))],  # go/NOTYPO
                        enable_all_sublanes_mask,
                    )
                    scf.YieldOp([])

                # undo reshape cast on scratch_local
                if not is_bf16:
                    mem_num = 128
                    new_scratch_shape = (
                        row_chunk_size * col_chunk_size // mem_num,
                        mem_num,
                    )
                    new_scratch_layout = ir.StridedLayoutAttr.get(
                        0, [mem_num, 1])
                    new_scratch_ref_ty = ir.MemRefType.get(
                        new_scratch_shape,
                        f32,  # bf16,
                        new_scratch_layout,
                        memory_space=memoryspace_tilespmem,
                    )

                    scratch_local = memref.reinterpret_cast(
                        new_scratch_ref_ty,
                        scratch_local,
                        [],
                        [],
                        [],
                        static_offsets=[0],
                        static_sizes=new_scratch_shape,
                        static_strides=[mem_num, 1],
                    )

                return scratch_out_local

            def fill_out_offset_tile(offset_tile_out_local,
                                     col_pos,
                                     row_pos=None):
                """Fills the offset tile for indirect DMA scatter for outputs (bf16).

        This function calculates the HBM offsets to scatter the reduced rows
        to. The offsets are calculated to correctly index into the output
        tensor in HBM, based on the row index of the reduction group
        `row_pos` and the current column chunk `col_pos`. The calculated
        offsets are stored in offset_tile_out_local, which is later used by
        tpu.enqueue_indirect_dma to scatter results from TileSpMem to HBM.

        Args:
          offset_tile_out_local: The destination memref in TileSpMem to store
            calculated offsets.
          col_pos: The index of the current column chunk being processed.
          row_pos: The starting row index for the reduction group in the
            output tensor. If None, computes offsets for prologue.

        Returns:
          The offset_tile_out_local memref filled with offsets for DMA scatter.
        """
                if row_pos is not None:

                    rest_row_pos = arith.remui(row_pos, const_lut(8))
                    rest_row_pos = arith.divui(rest_row_pos, const_lut(2))
                    tmp = arith.divui(row_pos, const_lut(8))
                    tmp = arith.muli(tmp, const_lut(4 * op.shape[1] // 128))
                    row_vec_offset = arith.addi(tmp, rest_row_pos)

                    row_vec_offset = arith.index_cast(i32, row_vec_offset)

                    row_vec_offset = vector.broadcast(_I32[row_chunk_size],
                                                      row_vec_offset)

                iota = arith.constant(
                    _I32[row_chunk_size],
                    ir.DenseIntElementsAttr.get(
                        array.array("i",
                                    [i * 4 for i in range(row_chunk_size)
                                     ]),  # 4 is key here for bf16
                        type=_I32[row_chunk_size],
                    ),
                )

                iota = arith.addi(
                    iota,
                    vector.broadcast(
                        _I32[row_chunk_size],
                        arith.muli(  # go/NOTYPO
                            arith.index_cast(i32, col_pos),
                            const_lut(col_chunk_size // 128 * 4, i32),
                        ),
                    ),
                )

                loop_over_idx = scf.ForOp(
                    lower_bound=const_lut(0),
                    upper_bound=const_lut(row_chunk_size // reduce_group_size),
                    step=const_lut(2),
                )
                loop_over_idx.attributes[
                    "sc.loop_unroll_factor"] = ir.IntegerAttr.get(
                        i32, loop_unroll_factor_2)
                if loop_parallel_access_2:
                    loop_over_idx.attributes[
                        "sc.parallel_access"] = ir.UnitAttr.get()
                with ir.InsertionPoint(loop_over_idx.body):
                    loop_i = loop_over_idx.induction_variable
                    start_vec = arith.addi(
                        vector.broadcast(_I32[row_chunk_size],
                                         arith.index_cast(i32, loop_i)),
                        iota,
                    )
                    loop_j = scf.ForOp(
                        const_lut(0),
                        const_lut((col_chunk_size // 128) // vreg_size),
                        const_lut(1),
                        iter_args=[start_vec],
                    )
                    loop_j.attributes[
                        "sc.loop_unroll_factor"] = ir.IntegerAttr.get(
                            i32, loop_unroll_factor_3)
                    if loop_parallel_access_3:
                        loop_j.attributes[
                            "sc.parallel_access"] = ir.UnitAttr.get()
                    with ir.InsertionPoint(loop_j.body):
                        vec = loop_j.inner_iter_args[0]

                        idx_local = arith.addi(
                            arith.muli(loop_i, const_lut(
                                (op.shape[1] // 128))),  # go/NOTYPO
                            arith.muli(  # go/NOTYPO
                                loop_j.induction_variable,
                                const_lut(vreg_size)),
                        )

                        if row_pos is not None:
                            local_vec = arith.addi(vec, row_vec_offset)
                        else:
                            local_vec = vec

                        tpu.store(
                            local_vec,
                            offset_tile_out_local,
                            [idx_local],
                            enable_all_sublanes_mask,
                        )
                        add_vec = vector.broadcast(
                            _I32[vreg_size], const_lut(4 * vreg_size, i32))
                        next_vec = arith.addi(vec, add_vec)
                        scf.YieldOp([next_vec])
                    scf.YieldOp([])

                rem_vreg = col_chunk_size / 128 % vreg_size / vreg_size

                last_full_vec = tpu.load(
                    _I32[vreg_size],
                    offset_tile_out_local,
                    [
                        const_lut(offset_sizes_out - int(vreg_size *
                                                         (1 + rem_vreg)))
                    ],
                    enable_all_sublanes_mask,
                )
                add_vec = vector.broadcast(_I32[vreg_size],
                                           const_lut(
                                               int(4 * rem_vreg * vreg_size),
                                               i32))  # 4 comes from bf16

                last_full_vec = arith.addi(last_full_vec, add_vec)
                tpu.store(
                    last_full_vec,
                    offset_tile_out_local,
                    [const_lut(offset_sizes_out - vreg_size)],
                    enable_all_sublanes_mask,
                )
                return offset_tile_out_local

            offset_sizes_out = ((row_chunk_size // reduce_group_size) *
                                (col_chunk_size // 128) // 2
                                )  # two because doing two rows at once
            offset_tile_out_0 = memref.alloca(
                ir.MemRefType.get(
                    shape=(offset_sizes_out, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            offset_tile_out_1 = memref.alloca(
                ir.MemRefType.get(
                    shape=(offset_sizes_out, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            if is_bf16:
                scratch_0 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size * 2, col_chunk_size),
                        element_type=bf16,  # f32,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )
                scratch_1 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size * 2, col_chunk_size),
                        element_type=bf16,  # f32,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )
                mem_num = 128 * 2
                new_scratch_shape = (
                    row_chunk_size * col_chunk_size * 2 // mem_num,
                    mem_num,
                )
                new_scratch_layout = ir.StridedLayoutAttr.get(0, [mem_num, 1])
                new_scratch_ref_ty = ir.MemRefType.get(
                    new_scratch_shape,
                    bf16,  # f32,
                    new_scratch_layout,
                    memory_space=memoryspace_tilespmem,
                )
            else:
                scratch_0 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size, col_chunk_size),
                        element_type=f32,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )
                scratch_1 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size, col_chunk_size),
                        element_type=f32,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )
                mem_num = 128
                new_scratch_shape = (
                    row_chunk_size * col_chunk_size // mem_num,
                    mem_num,
                )
                new_scratch_layout = ir.StridedLayoutAttr.get(0, [mem_num, 1])
                new_scratch_ref_ty = ir.MemRefType.get(
                    new_scratch_shape,
                    f32,
                    new_scratch_layout,
                    memory_space=memoryspace_tilespmem,
                )

            scratch_0 = memref.reinterpret_cast(new_scratch_ref_ty,
                                                scratch_0, [], [], [],
                                                static_offsets=[0],
                                                static_sizes=new_scratch_shape,
                                                static_strides=[mem_num, 1])

            scratch_1 = memref.reinterpret_cast(new_scratch_ref_ty,
                                                scratch_1, [], [], [],
                                                static_offsets=[0],
                                                static_sizes=new_scratch_shape,
                                                static_strides=[mem_num, 1])

            scratch_out_0 = memref.alloca(
                ir.MemRefType.get(
                    shape=(2, col_chunk_size),
                    element_type=bf16,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            scratch_out_1 = memref.alloca(
                ir.MemRefType.get(
                    shape=(2, col_chunk_size),
                    element_type=bf16,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            mem_num = 128 * 2
            new_scratch_shape = (2 * col_chunk_size // mem_num, mem_num)
            new_scratch_layout = ir.StridedLayoutAttr.get(0, [mem_num, 1])
            new_scratch_ref_ty = ir.MemRefType.get(
                new_scratch_shape,
                bf16,
                new_scratch_layout,
                memory_space=memoryspace_tilespmem,
            )

            scratch_out_0 = memref.reinterpret_cast(
                new_scratch_ref_ty,
                scratch_out_0, [], [], [],
                static_offsets=[0],
                static_sizes=new_scratch_shape,
                static_strides=[mem_num, 1])

            scratch_out_1 = memref.reinterpret_cast(
                new_scratch_ref_ty,
                scratch_out_1, [], [], [],
                static_offsets=[0],
                static_sizes=new_scratch_shape,
                static_strides=[mem_num, 1])

            new_input_shape = (
                idx.shape[0] // reduce_group_size * op.shape[1] // mem_num,
                mem_num,
            )
            new_input_ref_ty = ir.MemRefType.get(new_input_shape,
                                                 bf16,
                                                 memory_space=memoryspace_hbm)
            out_ref = tpu.reinterpret_cast(
                new_input_ref_ty,
                out_ref,
            )

            if topk_weights is not None:
                new_weights_shape = (topk_weights.size // 2, 2)
                new_weights_ref_ty = ir.MemRefType.get(
                    new_weights_shape, bf16, memory_space=memoryspace_hbm)
                weights_ref = tpu.reinterpret_cast(new_weights_ref_ty,
                                                   weights_ref)

            sflag_0 = tpu.sem_alloc(
                ir.MemRefType.get((),
                                  dma_semaphore_type,
                                  memory_space=memory_space_semaphore))
            sflag_1 = tpu.sem_alloc(
                ir.MemRefType.get((),
                                  dma_semaphore_type,
                                  memory_space=memory_space_semaphore))
            sflag_out_0 = tpu.sem_alloc(
                ir.MemRefType.get((),
                                  dma_semaphore_type,
                                  memory_space=memory_space_semaphore))
            sflag_out_1 = tpu.sem_alloc(
                ir.MemRefType.get((),
                                  dma_semaphore_type,
                                  memory_space=memory_space_semaphore))

            idx_tile_0 = memref.alloca(
                ir.MemRefType.get(
                    shape=(row_chunk_size, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            idx_tile_1 = memref.alloca(
                ir.MemRefType.get(
                    shape=(row_chunk_size, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            weights_tile_0 = None
            weights_tile_1 = None
            sflag_weights_0 = None
            sflag_weights_1 = None
            offset_tile_weights = None
            if topk_weights is not None:
                weights_tile_0 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size,
                               2),  # matches Offsets shape (16)
                        element_type=bf16,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )
                weights_tile_1 = memref.alloca(
                    ir.MemRefType.get(
                        shape=(row_chunk_size,
                               2),  # matches Offsets shape (16)
                        element_type=bf16,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )

                sflag_weights_0 = tpu.sem_alloc(
                    ir.MemRefType.get((),
                                      dma_semaphore_type,
                                      memory_space=memory_space_semaphore))
                sflag_weights_1 = tpu.sem_alloc(
                    ir.MemRefType.get((),
                                      dma_semaphore_type,
                                      memory_space=memory_space_semaphore))

                offset_tile_weights = memref.alloca(
                    ir.MemRefType.get(
                        shape=(vreg_size, ),
                        element_type=i32,
                        memory_space=memoryspace_tilespmem,
                    ),
                    [],
                    [],
                )

            offset_sizes = row_chunk_size * (col_chunk_size // 128)
            offset_tile_0 = memref.alloca(
                ir.MemRefType.get(
                    shape=(offset_sizes, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            offset_tile_1 = memref.alloca(
                ir.MemRefType.get(
                    shape=(offset_sizes, ),
                    element_type=i32,
                    memory_space=memoryspace_tilespmem,
                ),
                [],
                [],
            )

            global_chunk_to_process = arith.addi(
                arith.muli(current_sc_core, const_lut(num_sc_per_core,
                                                      i32)),  # go/NOTYPO
                current_local_core,
            )
            global_row_idx_start = arith.index_cast(
                index,
                arith.muli(  # go/NOTYPO
                    global_chunk_to_process,
                    const_lut(idx.shape[0] // num_sc, i32),
                ),
            )

            loop_col_chunk = scf.ForOp(
                const_lut(0),
                const_lut(op.shape[1] // col_chunk_size),
                const_lut(1),
            )
            with ir.InsertionPoint(loop_col_chunk.body):
                col_chunk_ij = loop_col_chunk.induction_variable
                offset_tile_out_0 = fill_out_offset_tile(
                    offset_tile_out_0, col_chunk_ij)

                # #############
                # # prologue
                # #############

                # setup stream for #0
                loop_row_chunk_idx = const_lut(0)

                base_idx_val = arith.index_cast(
                    i32, arith.addi(global_row_idx_start, loop_row_chunk_idx))
                base_idx_val = tpu.assume_multiple(base_idx_val, 8)
                source_slice = tpu.memref_slice(
                    ir.MemRefType.get((row_chunk_size, ),
                                      i32,
                                      memory_space=memoryspace_hbm),
                    idx_ref,
                    base_idx=[base_idx_val],
                    dynamic_sizes=[],
                )

                tpu.enqueue_dma(source_slice, idx_tile_0, sflag_0)
                tpu.wait_dma2(semaphore=sflag_0,
                              src=source_slice,
                              dst=idx_tile_0)

                if topk_weights is not None:
                    # The linear index is global_row_idx_start + loop_row_chunk_idx
                    lin_idx = arith.addi(global_row_idx_start,
                                         loop_row_chunk_idx)
                    parity_0 = load_weights(lin_idx, weights_tile_0,
                                            sflag_weights_0)
                    parity_1 = const_lut(0)
                else:
                    parity_0 = const_lut(0)
                    parity_1 = const_lut(0)

                offset_tile_0, idx_parity_0 = fill_load_offset_tile(
                    offset_tile_0, idx_tile_0, col_chunk_ij)

                if is_bf16:
                    mem_num = 128 * 2
                    new_input_shape = (op.shape[0] * op.shape[1] // mem_num,
                                       mem_num)
                    new_input_layout = ir.StridedLayoutAttr.get(
                        0, [mem_num, 1])
                    new_input_ref_ty = ir.MemRefType.get(
                        new_input_shape,
                        bf16,  # f32
                        new_input_layout,
                        memory_space=memoryspace_hbm,
                    )
                else:
                    mem_num = 128
                    new_input_shape = (op.shape[0] * op.shape[1] // mem_num,
                                       mem_num)
                    new_input_layout = ir.StridedLayoutAttr.get(
                        0, [mem_num, 1])
                    new_input_ref_ty = ir.MemRefType.get(
                        new_input_shape,
                        f32,
                        new_input_layout,
                        memory_space=memoryspace_hbm,
                    )
                op_ref = memref.reinterpret_cast(new_input_ref_ty,
                                                 op_ref, [], [], [],
                                                 static_offsets=[0],
                                                 static_sizes=new_input_shape,
                                                 static_strides=[mem_num, 1])

                tpu.enqueue_indirect_dma(
                    source=op_ref,
                    target=scratch_0,
                    offsets=offset_tile_0,
                    semaphore=sflag_0,
                )

                #############
                # corpus
                #############
                loop_over_row_chunks = scf.ForOp(
                    lower_bound=const_lut(row_chunk_size * 1),
                    upper_bound=const_lut(
                        row_chunk_size *
                        ((idx.shape[0] // num_sc // row_chunk_size) - 1)),
                    step=const_lut(row_chunk_size * 2),
                    iter_args=[
                        scratch_0,
                        scratch_out_0,
                        sflag_0,
                        sflag_out_0,
                        idx_tile_0,
                        offset_tile_0,
                        offset_tile_out_0,
                        scratch_1,
                        scratch_out_1,
                        sflag_1,
                        sflag_out_1,
                        idx_tile_1,
                        offset_tile_1,
                        offset_tile_out_1,
                        weights_tile_0
                        if topk_weights is not None else scratch_0,
                        weights_tile_1
                        if topk_weights is not None else scratch_0,
                        sflag_weights_0
                        if topk_weights is not None else sflag_0,
                        sflag_weights_1
                        if topk_weights is not None else sflag_0,
                        parity_0 if topk_weights is not None else const_lut(0),
                        parity_1 if topk_weights is not None else const_lut(0),
                        idx_parity_0,
                        idx_parity_0,  # idx_parity_1 is not initialized yet
                    ],
                )
                with ir.InsertionPoint(loop_over_row_chunks.body):
                    loop_row_chunk_idx = loop_over_row_chunks.induction_variable
                    (
                        scratch_0,
                        scratch_out_0,
                        sflag_0,
                        sflag_out_0,
                        idx_tile_0,
                        offset_tile_0,
                        offset_tile_out_0,
                        scratch_1,
                        scratch_out_1,
                        sflag_1,
                        sflag_out_1,
                        idx_tile_1,
                        offset_tile_1,
                        offset_tile_out_1,
                        weights_tile_0,
                        weights_tile_1,
                        sflag_weights_0,
                        sflag_weights_1,
                        parity_0,
                        parity_1,
                        idx_parity_0,
                        idx_parity_1,
                    ) = loop_over_row_chunks.inner_iter_args

                    # setup stream for #1
                    base_idx_val = arith.index_cast(
                        i32,
                        arith.addi(
                            global_row_idx_start,
                            arith.addi(loop_row_chunk_idx, const_lut(0)),
                        ),
                    )
                    base_idx_val = tpu.assume_multiple(base_idx_val, 8)
                    source_slice = tpu.memref_slice(
                        ir.MemRefType.get((row_chunk_size, ),
                                          i32,
                                          memory_space=memoryspace_hbm),
                        idx_ref,
                        base_idx=[base_idx_val],
                        dynamic_sizes=[],
                    )

                    tpu.enqueue_dma(source_slice, idx_tile_1, sflag_1)
                    tpu.wait_dma2(semaphore=sflag_1,
                                  src=source_slice,
                                  dst=idx_tile_1)

                    if topk_weights is not None:
                        # The linear index is global_row_idx_start + loop_row_chunk_idx
                        lin_idx = arith.addi(global_row_idx_start,
                                             loop_row_chunk_idx)
                        parity_1 = load_weights(lin_idx, weights_tile_1,
                                                sflag_weights_1)

                    offset_tile_1, idx_parity_1 = fill_load_offset_tile(
                        offset_tile_1, idx_tile_1, col_chunk_ij)

                    tpu.enqueue_indirect_dma(
                        source=op_ref,
                        target=scratch_1,
                        offsets=offset_tile_1,
                        semaphore=sflag_1,
                    )

                    # wait stream #0
                    tpu.wait_indirect_dma(semaphore=sflag_0,
                                          src=op_ref,
                                          dst=scratch_0)

                    # process #0
                    scratch_out_0 = perform_add(
                        scratch_0,
                        scratch_out_0,
                        idx_parity_0,
                        weights_local=weights_tile_0
                        if topk_weights is not None else None,
                        parity=parity_0 if topk_weights is not None else None,
                    )

                    offset_tile_out_0 = fill_out_offset_tile(
                        offset_tile_out_0,
                        col_chunk_ij,
                        arith.divui(
                            arith.addi(
                                global_row_idx_start,
                                arith.subi(loop_row_chunk_idx,
                                           const_lut(row_chunk_size)),
                            ),
                            const_lut(reduce_group_size),
                        ),
                    )
                    tpu.enqueue_indirect_dma(
                        source=scratch_out_0,
                        target=out_ref,
                        offsets=offset_tile_out_0,
                        semaphore=sflag_out_0,
                    )

                    if topk_weights is not None:
                        # The linear index is:
                        # global_row_idx_start + loop_row_chunk_idx + row_chunk_size
                        lin_idx = arith.addi(
                            global_row_idx_start,
                            arith.addi(loop_row_chunk_idx,
                                       const_lut(row_chunk_size)),
                        )
                        parity_0 = load_weights(lin_idx, weights_tile_0,
                                                sflag_weights_0)

                    base_idx_val = arith.index_cast(
                        i32,
                        arith.addi(
                            global_row_idx_start,
                            arith.addi(loop_row_chunk_idx,
                                       const_lut(row_chunk_size)),
                        ),
                    )
                    base_idx_val = tpu.assume_multiple(base_idx_val, 8)
                    source_slice = tpu.memref_slice(
                        ir.MemRefType.get((row_chunk_size, ),
                                          i32,
                                          memory_space=memoryspace_hbm),
                        idx_ref,
                        base_idx=[base_idx_val],
                        dynamic_sizes=[],
                    )
                    tpu.enqueue_dma(source_slice, idx_tile_0, sflag_0)
                    tpu.wait_dma2(semaphore=sflag_0,
                                  src=source_slice,
                                  dst=idx_tile_0)

                    offset_tile_0, idx_parity_0 = fill_load_offset_tile(
                        offset_tile_0, idx_tile_0, col_chunk_ij)

                    tpu.enqueue_indirect_dma(
                        source=op_ref,
                        target=scratch_0,
                        offsets=offset_tile_0,
                        semaphore=sflag_0,
                    )

                    # wait stream #1
                    tpu.wait_indirect_dma(semaphore=sflag_1,
                                          src=op_ref,
                                          dst=scratch_1)

                    # # process #1
                    scratch_out_1 = perform_add(
                        scratch_1,
                        scratch_out_1,
                        idx_parity_1,
                        weights_local=weights_tile_1
                        if topk_weights is not None else None,
                        parity=parity_1 if topk_weights is not None else None,
                    )

                    offset_tile_out_1 = fill_out_offset_tile(
                        offset_tile_out_1,
                        col_chunk_ij,
                        arith.divui(
                            arith.addi(global_row_idx_start,
                                       loop_row_chunk_idx),
                            const_lut(reduce_group_size),
                        ),
                    )
                    tpu.enqueue_indirect_dma(
                        source=scratch_out_1,
                        target=out_ref,
                        offsets=offset_tile_out_1,
                        semaphore=sflag_out_1,
                    )

                    # wait stream #0 out
                    tpu.wait_indirect_dma(semaphore=sflag_out_0,
                                          src=scratch_out_0,
                                          dst=out_ref)
                    # wait stream #1 out
                    tpu.wait_indirect_dma(semaphore=sflag_out_1,
                                          src=scratch_out_1,
                                          dst=out_ref)

                    # return #0 and #1
                    scf.YieldOp([
                        scratch_0,
                        scratch_out_0,
                        sflag_0,
                        sflag_out_0,
                        idx_tile_0,
                        offset_tile_0,
                        offset_tile_out_0,
                        scratch_1,
                        scratch_out_1,
                        sflag_1,
                        sflag_out_1,
                        idx_tile_1,
                        offset_tile_1,
                        offset_tile_out_1,
                        weights_tile_0
                        if topk_weights is not None else scratch_0,
                        weights_tile_1
                        if topk_weights is not None else scratch_0,
                        sflag_weights_0
                        if topk_weights is not None else sflag_0,
                        sflag_weights_1
                        if topk_weights is not None else sflag_0,
                        parity_0 if topk_weights is not None else const_lut(0),
                        parity_1 if topk_weights is not None else const_lut(0),
                        idx_parity_0,
                        idx_parity_1,
                    ])

                #############
                # epilogue
                #############
                (
                    scratch_0,
                    scratch_out_0,
                    sflag_0,
                    sflag_out_0,
                    _,
                    _,
                    offset_tile_out_0,
                    scratch_1,
                    scratch_out_1,
                    sflag_1,
                    sflag_out_1,
                    idx_tile_1,
                    offset_tile_1,
                    offset_tile_out_1,
                    weights_tile_0,
                    weights_tile_1,
                    _,
                    sflag_weights_1,
                    parity_0,
                    parity_1,
                    idx_parity_0,
                    idx_parity_1,
                ) = loop_over_row_chunks.results_

                epi_idx_loc = row_chunk_size * (
                    (idx.shape[0] // num_sc // row_chunk_size) - 1)
                add_f = arith.divui(global_row_idx_start,
                                    const_lut(reduce_group_size))

                # setup stream for #1
                base_idx_val = arith.index_cast(
                    i32,
                    arith.addi(global_row_idx_start, const_lut(epi_idx_loc)))
                base_idx_val = tpu.assume_multiple(base_idx_val, 8)
                source_slice = tpu.memref_slice(
                    ir.MemRefType.get((row_chunk_size, ),
                                      i32,
                                      memory_space=memoryspace_hbm),
                    idx_ref,
                    base_idx=[base_idx_val],
                    dynamic_sizes=[],
                )
                tpu.enqueue_dma(source_slice, idx_tile_1, sflag_1)
                tpu.wait_dma2(semaphore=sflag_1,
                              src=source_slice,
                              dst=idx_tile_1)

                if topk_weights is not None:
                    lin_idx = arith.addi(global_row_idx_start,
                                         const_lut(epi_idx_loc))
                    parity_1 = load_weights(lin_idx, weights_tile_1,
                                            sflag_weights_1)

                offset_tile_1, idx_parity_1 = fill_load_offset_tile(
                    offset_tile_1, idx_tile_1, col_chunk_ij)

                tpu.enqueue_indirect_dma(
                    source=op_ref,
                    target=scratch_1,
                    offsets=offset_tile_1,
                    semaphore=sflag_1,
                )

                # wait stream #0
                tpu.wait_indirect_dma(semaphore=sflag_0,
                                      src=op_ref,
                                      dst=scratch_0)

                # process #0
                scratch_out_0 = perform_add(
                    scratch_0,
                    scratch_out_0,
                    idx_parity_0,
                    weights_local=weights_tile_0
                    if topk_weights is not None else None,
                    parity=parity_0 if topk_weights is not None else None,
                )

                offset_tile_out_0 = fill_out_offset_tile(
                    offset_tile_out_0,
                    col_chunk_ij,
                    arith.addi(
                        add_f,
                        const_lut((idx.shape[0] // num_sc //
                                   reduce_group_size) - 4),
                    ),
                )
                tpu.enqueue_indirect_dma(
                    source=scratch_out_0,
                    target=out_ref,
                    offsets=offset_tile_out_0,
                    semaphore=sflag_out_0,
                )

                # wait stream #1
                tpu.wait_indirect_dma(semaphore=sflag_1,
                                      src=op_ref,
                                      dst=scratch_1)

                # process #1
                scratch_out_1 = perform_add(
                    scratch_1,
                    scratch_out_1,
                    idx_parity_1,
                    weights_local=weights_tile_1
                    if topk_weights is not None else None,
                    parity=parity_1 if topk_weights is not None else None,
                )

                offset_tile_out_1 = fill_out_offset_tile(
                    offset_tile_out_1,
                    col_chunk_ij,
                    arith.addi(
                        add_f,
                        const_lut((idx.shape[0] // num_sc //
                                   reduce_group_size) - 2),
                    ),
                )
                tpu.enqueue_indirect_dma(
                    source=scratch_out_1,
                    target=out_ref,
                    offsets=offset_tile_out_1,
                    semaphore=sflag_out_1,
                )
                tpu.wait_indirect_dma(semaphore=sflag_out_0,
                                      src=scratch_out_0,
                                      dst=out_ref)
                tpu.wait_indirect_dma(semaphore=sflag_out_1,
                                      src=scratch_out_1,
                                      dst=out_ref)

                scf.YieldOp([])

        input_types = [
            i32,
            i32,
            ir.MemRefType.get(
                idx.shape,
                i32,
                memory_space=memoryspace_hbm,
            ),
            ir.MemRefType.get(
                op.shape,
                bf16 if is_bf16 else f32,
                memory_space=memoryspace_hbm,
            ),
            ir.MemRefType.get(
                (idx.shape[0] // reduce_group_size, op.shape[1]),
                bf16,
                memory_space=memoryspace_hbm,
            ),
        ]
        if topk_weights is not None:
            input_types.insert(
                4,
                ir.MemRefType.get(
                    topk_weights.shape,
                    bf16,
                    memory_space=memoryspace_hbm,
                ),
            )

        # Configure the wrappers
        if topk_weights is not None:

            @func.FuncOp.from_py_func(*input_types, name="main")
            def kernel_main(
                current_sc_core,
                current_local_core,
                idx_ref,
                op_ref,
                weights_ref,
                out_ref,
                func_op,
            ):
                return _kernel_impl(
                    current_sc_core,
                    current_local_core,
                    idx_ref,
                    op_ref,
                    weights_ref,
                    out_ref,
                    func_op,
                )

        else:

            @func.FuncOp.from_py_func(*input_types, name="main")
            def kernel_main(
                current_sc_core,
                current_local_core,
                idx_ref,
                op_ref,
                out_ref,
                func_op,
            ):
                return _kernel_impl(
                    current_sc_core,
                    current_local_core,
                    idx_ref,
                    op_ref,
                    None,
                    out_ref,
                    func_op,
                )

        # Configure the Mosaic iteration space
        f = kernel_main.func_op
        f.attributes["iteration_bounds"] = ir.DenseI64ArrayAttr.get(
            [used_sc_cores, num_sc_per_core])
        f.attributes["dimension_semantics"] = ir.ArrayAttr.get([
            ir.Attribute.parse("#tpu.dimension_semantics<core_parallel>"),
            ir.Attribute.parse("#tpu.dimension_semantics<parallel>"),
        ])
        args_attributes = [
            ir.DictAttr.get({}),  # i
            ir.DictAttr.get({}),  # j
            ir.DictAttr.get({"sc.persistent": ir.UnitAttr.get()}),  # idx
            ir.DictAttr.get({"sc.persistent": ir.UnitAttr.get()}),  # op
            ir.DictAttr.get({"sc.persistent": ir.UnitAttr.get()}),  # out
        ]
        window_params = [
            ir.DictAttr.get({
                "transform_indices":
                ir.Attribute.parse("affine_map<(n,m) -> (0)>"),
            }),
            ir.DictAttr.get({
                "transform_indices":
                ir.Attribute.parse("affine_map<(n,m) -> (0,0)>"),
            }),
            ir.DictAttr.get({
                "transform_indices":
                ir.Attribute.parse("affine_map<(n,m) -> (0,0)>"),
            }),
        ]

        if topk_weights is not None:
            # Insert weights before output - we append here because its the same
            # attribute as output.
            args_attributes.append(
                ir.DictAttr.get({"sc.persistent": ir.UnitAttr.get()}))
            window_params.append(
                ir.DictAttr.get({
                    "transform_indices":
                    ir.Attribute.parse("affine_map<(n,m) -> (0,0)>"),
                }), )

        f.attributes["window_params"] = ir.ArrayAttr.get(window_params)
        f.arg_attrs = args_attributes
        f.attributes["tpu.core_type"] = ir.Attribute.parse(
            "#tpu.core_type<sc_vector_subcore>")
        assert f.verify(), f
        m = ir.Module.create()
        m.body.append(f)
        ir.SymbolTable(m.operation).insert(f)

        return mosaic.as_tpu_kernel(
            m,
            out_type=core.ShapedArray(
                (idx.shape[0] // reduce_group_size, op.shape[1]),
                jnp.bfloat16),
        )(*([
            idx,
            op,
            topk_weights,
        ] if topk_weights is not None else [
            idx,
            op,
        ]))



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import NamedSharding
from jax.sharding import PartitionSpec as P

from tpu_inference.layers.common.moe import MoEBackend, moe_apply
from tpu_inference.layers.common.process_weights.moe_weights import (
    FusedMoEWeights, UnfusedMoEWeights)
from tpu_inference.layers.common.quantization import unquantized as jax_common
from tpu_inference.layers.common.quantization.configs import QuantLinearConfig
from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.linear import JaxEinsum
from tpu_inference.layers.jax.moe.moe import JaxMoE
from tpu_inference.layers.jax.quantization import QuantizeMethodBase
from tpu_inference.layers.jax.quantization.configs import QuantizationConfig
from tpu_inference.models.jax.utils.weight_utils import shard_put


class UnquantizedLinearMethod(QuantizeMethodBase,
                              jax_common.UnquantizedLinearMethod):
    """Unquantized method for JAX Linear layer.
    """

    def apply_jax(self, layer: JaxModule, x: jax.Array) -> jax.Array:
        assert isinstance(layer, JaxEinsum)

        with jax.named_scope(layer._get_name()):
            if self.linear_config.fuse_matmuls:
                out = self._apply_fused(
                    x,
                    layer.weight.value,
                    layer.bias.value if layer.bias else None,
                    einsum_str=layer.einsum_str)
            else:
                raise NotImplementedError(
                    "Non-fused matmuls not implemented yet.")

        return out


class UnquantizedFusedMoEMethod(QuantizeMethodBase):
    """
    Unquantized method for JAXMoE layer.

    TODO (jacobplatin): support weight loading -- currently, model-dependent.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extra_backend_kwargs = {}

    def process_weights_after_loading(self, layer: JaxMoE, *args,
                                      **kwargs) -> bool:
        """
        Process weights after loading.

        Please see https://github.com/vllm-project/tpu-inference/blob/bb1a88/tpu_inference/layers/common/moe.py#L39
        for more information on the expected weights per MoE backend.

        Args:
            layer: The layer to process.
        """
        if layer.moe_backend == MoEBackend.FUSED_MOE:
            if layer.edf_sharding:
                e2df_sharding = (layer.edf_sharding[0], None,
                                 layer.edf_sharding[1], layer.edf_sharding[2])
            # fuse the weights into w13: [Gate, Up]
            w_gate = layer.kernel_gating_EDF.value
            w_up = layer.kernel_up_proj_EDF.value

            # stack to create a 4d array
            w13_val = jnp.stack([w_gate, w_up], axis=1)

            layer.kernel_gating_upproj_E2DF = nnx.Param(
                shard_put(w13_val, shardings=e2df_sharding))

            del layer.kernel_gating_EDF
            del layer.kernel_up_proj_EDF

            ep_axis_name = layer.efd_sharding[0]

            self.extra_backend_kwargs = {
                "ep_axis_name": ep_axis_name,
                "bt": 32,
                "bf": 512,
                "bd1": 512,
                "bd2": 512,
                "btc": 64,
                "bfc": 256,
                "bd1c": 256,
                "bd2c": 256,
            }

        elif layer.moe_backend in [MoEBackend.GMM_EP, MoEBackend.GMM_TP]:
            if any(
                    any(w is None for w in param._weights_to_load)
                    for param in [
                        layer.kernel_gating_EDF, layer.kernel_up_proj_EDF,
                        layer.kernel_down_proj_EFD
                    ]):
                return False
            w_gate = layer.kernel_gating_EDF.get_value()
            w_up = layer.kernel_up_proj_EDF.get_value()

            # Fuse the weights into w13: [Gate, Up]
            w13_val = jnp.concatenate([w_gate, w_up], axis=1)

            weights = jax_common.process_unquantized_moe_weights(
                mesh=jax.sharding.get_mesh(),
                moe_backend=layer.moe_backend,
                activation=layer.activation,
                w13_weight=w13_val,
                w13_bias=None,
                w2_weight=layer.kernel_down_proj_EFD.get_value(),
                w2_bias=None,
            )

            # TODO (jacobplatin): we probably want to make the sharding configurable
            layer.kernel_gating_upproj_EDF = nnx.Param(weights.w13_weight)
            layer.kernel_down_proj_EFD = nnx.Param(weights.w2_weight)

            del layer.kernel_gating_EDF
            del layer.kernel_up_proj_EDF
            del weights
            del w13_val

        return True

    def apply_jax(self, layer: JaxMoE, x: jax.Array, *,
                  router_logits: jax.Array) -> jax.Array:
        """Forward pass for MoE layer.
        Args:
            layer: The MoE layer to apply.
            x: The input activations to the MoE layer, of shape [seq_len, hidden_size].
            router_logits: The routing logits for the MoE layer, of shape [seq_len, num_experts].
        """
        x_TD = jnp.asarray(x, layer.dtype)
        x_TD = jax.lax.with_sharding_constraint(
            x_TD, NamedSharding(layer.mesh, P(*layer.activation_ffw_td)))

        # Fused weight backends
        if layer.moe_backend in MoEBackend.fused_moe_backends():
            # router_logits is of shape TE, only 1D in this case

            w13_weight = layer.kernel_gating_upproj_E2DF.value if layer.moe_backend == MoEBackend.FUSED_MOE else layer.kernel_gating_upproj_EDF.value
            w2_weight = layer.kernel_down_proj_EFD.value
            # TODO (jacobplatin/bzgoogle): we should support bias
            weights = FusedMoEWeights(
                w13_weight=w13_weight,
                w13_weight_scale=None,
                w13_bias=None,
                w2_weight=w2_weight,
                w2_weight_scale=None,
                w2_bias=None,
            )
        elif layer.moe_backend in [
                MoEBackend.DENSE_MAT, MoEBackend.MEGABLX_GMM
        ]:
            # router_logits is composed of weights_TX and indices_TX, so 2D in this case
            # TODO (jacobplatin/bzgoogle): we should support bias
            weights = UnfusedMoEWeights(
                w1_weight=layer.kernel_gating_EDF.value,
                w1_weight_scale=None,
                w1_bias=None,
                w2_weight=layer.kernel_up_proj_EDF.value,
                w2_weight_scale=None,
                w2_bias=None,
                w3_weight=layer.kernel_down_proj_EFD.value,
                w3_weight_scale=None,
                w3_bias=None,
            )

        else:
            raise ValueError(f"Unsupported moe backend {layer.moe_backend}")
        return moe_apply(layer, x_TD, router_logits, weights,
                         layer.moe_backend, layer.mesh,
                         self.extra_backend_kwargs)


class UnquantizedConfig(QuantizationConfig):

    def get_quant_method(self, layer: JaxModule,
                         prefix: str) -> Optional[QuantizeMethodBase]:
        if isinstance(layer, JaxEinsum):
            # Derive output's last dim from the einsum string.
            einsum_str = layer.einsum_str.replace(" ", "")
            _, w_axis = einsum_str.split("->")[0].split(",")
            last_out_char = einsum_str.split("->")[1][-1]
            out_size = layer.kernel_shape[w_axis.index(last_out_char)]

            linear_config = QuantLinearConfig(enable_sp=False,
                                              output_sizes=[out_size])
            return UnquantizedLinearMethod(linear_config)
        if isinstance(layer, JaxMoE):
            return UnquantizedFusedMoEMethod()
        return None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A TPU Kernel for N:M structured sparse matmul (SPMM).

Note: this kernel provides software emulation of N:M SPMM on TPU. You will 
only get performance benefits if the workload is memory bound.

The N:M sparsity means at most N nonzeros in every set of M values where each
value are separated in a given stride. For example, 1:4 with stride 2 sparse
matrix with shape (8,) could be [1, 0, 0, 1, 0, 0, 0, 0]. There is only one
nonzero among indices 0, 2, 4, 6 and one nonzero among indices 1, 3, 5, 7.

This SPMM kernel recovers nonzeros and metadata (compressed nonzeros indices)
to sparse matrix in VMEM and matmul with the dense matrix.

We expect to support
  - any N:M sparsity as long as M <= 16;
  - data type: f32, bf16, int8;
  - either LHS or RHS is sparse matrix but not both;
  - either contracting or non-contracting dimension is sparse;
  - either RHS is transposed or not.
  - pass a default value for the masked out data.

Example of activation @ sparse_weight:
```
key = random.PRNGKey(1234)
mask = gen_sparse_mask(
    key, weight.shape, sparsity, stride=stride, sparse_dim=sparse_dim)

sparse_weight = Sparsifier(
    weight, mask, sparsity=sparsity, sparse_dim=sparse_dim, stride=stride
)

result = structured_spmm(
    sparsity,
    sparse_weight.nonzeros,
    sparse_weight.metadata,
    activation,
    sparse_dim,
    rhs_sparse,
    rhs_transpose,
    stride=stride,
    block_m,
    block_k,
    block_n,
)
```

Further improvements:
- TODO(jevinjiang): Consider pass Sparsifier object as input.
- TODO(jevinjiang): Consider use one decompress function
- TODO(jevinjiang): Support int4 dtype in Mosaic.
- TODO(jevinjiang): Support subelement mask in Mosaic.
- TODO(jevinjiang): Emulate pack/unpack for unsupported HW.
- TODO(jevinjiang): Add int2 dtype in JAX, so let astype do pack/unpack on HW.
"""

import functools
import math
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax import random
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu


class Sparsifier:
    """The sparsifier for structured sparse matrix.

  The sparsifier encodes sparse matrix to nonzeros and metadata.

  Attributes:
    shape: the shape of original data. Only used for decode.
    dtype: the dtype of original data. Only used for decode
    sparsity: the N:M structured sparsity.
    stride: each value in N:M structured are separated in stride.
    sparse_dim: the sparse dimension.
    nonzeros: the nonzeros gathered from original data.
    metadata: the compressed nonzeros indices.
  """

    def __init__(
        self,
        data: jax.Array,
        mask: jax.Array,
        *,
        sparsity: tuple[int, int],
        sparse_dim: int,
        stride: int,
    ):
        self.shape = data.shape
        self.dtype = data.dtype
        self.sparsity = sparsity
        self.stride = stride
        self.sparse_dim = sparse_dim
        # TODO(jevinjiang) validate inputs.
        self.nonzeros, self.metadata = self.encode(data, mask)

    def encode(self, data, mask):
        """Encode data to nonzeros and metadata based on the mask."""
        if self.sparse_dim == 0:
            data = data.transpose()
            mask = mask.transpose()
        x, y = self.sparsity
        h, w = data.shape
        data = data.reshape(-1, y, self.stride).transpose(0, 2, 1)
        mask = mask.reshape(-1, y, self.stride).transpose(0, 2, 1)
        _, nz_idx = jax.lax.top_k(mask, x)
        # Let CPU do the gather work.
        data = jax.device_get(data)
        nz = data[
            np.arange(nz_idx.shape[0])[:, None, None],
            np.arange(nz_idx.shape[1])[:, None],
            nz_idx,
        ]
        nz = jax.device_put(nz)
        nz_idx = nz_idx.transpose(2, 0, 1).reshape(x, h, w // y)
        nz = nz.transpose(2, 0, 1).reshape(x, h, w // y)
        if self.sparse_dim == 0:
            nz = nz.transpose(0, 2, 1)
            nz_idx = nz_idx.transpose(0, 2, 1)
        metadata = self.compress(nz_idx, next_log2(y))
        return nz, metadata

    def decode(self):
        """Decode nonzeros and metadata to sparse matrix and mask."""
        _, y = self.sparsity
        nz = self.nonzeros
        nz_idx = self.decompress(self.metadata, next_log2(y))
        h, w = self.shape
        if self.sparse_dim == 0:
            h, w = w, h
            nz = nz.transpose(0, 2, 1)
            nz_idx = nz_idx.transpose(0, 2, 1)
        nz = nz.transpose(1, 2, 0)
        nz_idx = nz_idx.transpose(1, 2, 0)
        # Let CPU do the scatter work.
        data = np.zeros((h, w // y, y), dtype=self.dtype)
        mask = np.zeros((h, w // y, y), dtype=jnp.int32)
        data[
            np.arange(nz_idx.shape[0])[:, None, None],
            np.arange(nz_idx.shape[1])[:, None],
            nz_idx,
        ] = nz
        mask[
            np.arange(nz_idx.shape[0])[:, None, None],
            np.arange(nz_idx.shape[1])[:, None],
            nz_idx,
        ] = 1
        data = data.reshape(-1, self.stride, y).transpose(0, 2,
                                                          1).reshape(h, w)
        mask = mask.reshape(-1, self.stride, y).transpose(0, 2,
                                                          1).reshape(h, w)
        data = jax.device_put(data)
        mask = jax.device_put(mask)
        if self.sparse_dim == 0:
            data = data.transpose()
            mask = mask.transpose()
        return data, mask

    def compress(self, data, bitwidth):
        """Pack the rows of data based on the element's bitwidth."""
        if bitwidth >= 32:
            return data
        d0, d1, d2 = data.shape
        bitwidth = next_pow2(bitwidth)
        packing = 32 // bitwidth
        assert d1 % packing == 0
        comp = np.zeros((d0, d1 // packing, d2), dtype=jnp.int32)
        for i in range(d1):
            shift = i % packing * bitwidth
            comp_i = i // packing
            comp[:, comp_i, :] |= data[:, i, :] << shift
        comp = jax.device_put(comp)
        return comp

    def decompress(self, data, bitwidth):
        """Unpack the rows of data based on the element's bitwidth."""
        if bitwidth >= 32:
            return data
        bitwidth = next_pow2(bitwidth)
        packing = 32 // bitwidth
        d0, d1, d2 = data.shape
        d1 *= packing
        result = np.zeros((d0, d1, d2), dtype=jnp.int32)
        for i in range(d1):
            shift = i % packing * bitwidth
            comp_i = i // packing
            result[:, i, :] = (data[:, comp_i, :] >> shift) & (2**bitwidth - 1)
        result = jax.device_put(result)
        return result


def next_log2(x):
    return math.ceil(math.log2(x))


def next_pow2(x):
    return 2**next_log2(x)


def gen_sparse_mask(
    key: Any,
    shape: jax.Array | tuple[int, int],
    sparsity: tuple[int, int],
    *,
    sparse_dim: int,
    stride: int = 1,
) -> jax.Array:
    """Generates a mask with N:M sparsity on a given dim.

  Args:
    key: random key.
    shape: the shape of the mask.
    sparsity: the N:M structured sparsity.
    sparse_dim: the sparse dimension.
    stride: each value in N:M structured are separated in stride.

  Returns:
    A mask with given shape and bool data.

  Raises:
    ValueError: An error occurred passing invalid input.
  """
    if sparse_dim not in (0, 1):
        raise ValueError(f"sparse_dim must be 0 or 1, got {sparse_dim}")
    if len(shape) != 2:
        raise ValueError(f"shape must be 2D, got {shape}")
    if len(sparsity) != 2:
        raise ValueError(
            f"expected 2 values in sparsity, but got {len(sparsity)}")
    h, w = shape
    x, y = sparsity
    if sparse_dim == 0:
        return gen_sparse_mask(key, (w, h),
                               sparsity,
                               sparse_dim=1,
                               stride=stride).transpose()
    if not (y > x > 0 and h > 0 and w > 0 and stride > 0):
        raise ValueError("invalid inputs")
    if w % (y * stride) != 0:
        raise ValueError(f"expected width {w} is a multiple of {y * stride}")
    nd_idx = jax.lax.broadcasted_iota(jnp.int32, (h * w // y, y), 1)
    mask = random.permutation(key, nd_idx, -1, independent=True) < x
    return mask.reshape(-1, stride, y).transpose(0, 2, 1).reshape(h, w)


def _get_dot_general_dim_nums(transposed: bool):
    return (((1, ), (1, )), ((), ())) if transposed else (((1, ), (0, )), ((),
                                                                           ()))


def _decompress_metadata(md_tile_ref, packing):
    """Decompress metadata. Expected to use in a Pallas kernel."""
    x, h, w = md_tile_ref.shape
    bitwidth = 32 // packing
    decompressed_md = []
    for i in range(h):
        unpacked_md = jnp.broadcast_to(md_tile_ref[:, pl.ds(i, 1), :],
                                       (x, packing, w))
        shift = jax.lax.broadcasted_iota(jnp.int32, unpacked_md.shape,
                                         1) * bitwidth
        unpacked_md = jax.lax.bitwise_and(
            jax.lax.shift_right_logical(unpacked_md, shift),
            jnp.broadcast_to(2**bitwidth - 1, unpacked_md.shape),
        )
        decompressed_md.append(unpacked_md)
    return jnp.concatenate(decompressed_md, axis=1)


def _decompress_nonzeros(
    sparsity,
    nonzeros,
    nonzeros_idx,
    sparse_dim,
    stride,
    default_value,
):
    """Decompress nonzeros. Expected to use in a Pallas kernel."""
    x, y = sparsity
    target_size = nonzeros.shape[1 + sparse_dim] * y
    tiles: Any = [None] * (target_size // stride)
    for i in range(y):
        lhs_tile_part = default_value
        for xi in range(x):
            lhs_tile_part = jnp.where(nonzeros_idx[xi] == i, nonzeros[xi],
                                      lhs_tile_part)
        for j in range(target_size // (y * stride)):
            if sparse_dim == 1:
                tiles[j * y + i] = lhs_tile_part[:,
                                                 j * stride:(j + 1) * stride]
            else:
                tiles[j * y + i] = lhs_tile_part[j * stride:(j + 1) *
                                                 stride, :]
    return jnp.concatenate(tiles, axis=sparse_dim).astype(nonzeros.dtype)


def _get_metadata_packing(sparsity_base):
    md_bitwidth = next_pow2(next_log2(sparsity_base))
    return 32 // md_bitwidth


def _get_dim_mapping(rhs_sparse: bool, rhs_transpose: bool):
    """The mapping to (m, n, k) dimensions."""
    m, n, k = (0, 1, 2)
    if rhs_sparse:
        return ((n, k), (m, k)) if rhs_transpose else ((k, n), (m, k))
    else:
        return ((m, k), (n, k)) if rhs_transpose else ((m, k), (k, n))


def _get_in_sepcs(
    sparsity: tuple[int, int],
    sparse_dim: int,
    rhs_sparse: bool,
    rhs_transpose: bool,
    block_m: int,
    block_k: int,
    block_n: int,
):
    """Get in_specs of structured SPMM."""
    x, y = sparsity
    md_packing = _get_metadata_packing(y)
    s_dims, d_dims = _get_dim_mapping(rhs_sparse, rhs_transpose)
    block_sz = (block_m, block_n, block_k)
    nz_shape = [x, block_sz[s_dims[0]], block_sz[s_dims[1]]]
    nz_shape[sparse_dim + 1] //= y
    md_shape = nz_shape.copy()
    md_shape[-2] //= md_packing
    nz_shape = tuple(nz_shape)
    md_shape = tuple(md_shape)
    mat_shape = (block_sz[d_dims[0]], block_sz[d_dims[1]])

    def get_sparse_index_map(i, j, k):
        indices = (i, j, k)
        return (0, indices[s_dims[0]], indices[s_dims[1]])

    def get_dense_index_map(i, j, k):
        indices = (i, j, k)
        return (indices[d_dims[0]], indices[d_dims[1]])

    return [
        pl.BlockSpec(nz_shape, get_sparse_index_map),
        pl.BlockSpec(md_shape, get_sparse_index_map),
        pl.BlockSpec(mat_shape, get_dense_index_map)
    ]


def _infer_out_dtype(ty1: jnp.dtype, ty2: jnp.dtype):
    """Infer matmul output dtype."""
    if ty1 == ty2:
        return ty1
    if jnp.issubdtype(ty1, jnp.integer) == jnp.issubdtype(ty2, jnp.integer):
        return ty1 if jnp.dtype(ty1).itemsize > jnp.dtype(
            ty2).itemsize else ty2
    return jnp.float32


def _verify(
    sparsity: tuple[int, int],
    nz: jax.Array,
    md: jax.Array,
    mat: jax.Array,
    *,
    sparse_dim: int,
    rhs_sparse: bool,
    rhs_transpose: bool,
    stride: int,
    block_m: int,
    block_k: int,
    block_n: int,
):
    """Validate inputs of structured SPMM and return original dim size m, n, k."""
    ##########################################################################
    # Verify sparsity.
    ##########################################################################
    x, y = sparsity
    if not (y > x > 0):
        raise ValueError(f"Invalid sparsity: {sparsity}")
    if y > 16:
        # TODO(jevinjiang): this requires padding because one compressed row is
        # decompressed to < 8 sublanes.
        raise ValueError("Not implemented: when sparsity base is > 16")
    if sparse_dim not in (0, 1):
        raise ValueError(f"Expected sparse_dim is 0 or 1, got {sparse_dim}")
    ##########################################################################
    # Verify shape and dtype.
    ##########################################################################
    if len(nz.shape) != 3:
        raise ValueError(f"Expected nonzeros' rank is 3, got {len(nz.shape)}")
    if len(md.shape) != 3:
        raise ValueError(f"Expected metadata's rank is 3, got {len(md.shape)}")
    if len(mat.shape) != 2:
        raise ValueError(f"Expected matrix's rank is 2, got {len(mat.shape)}")
    if md.dtype != jnp.int32:
        raise ValueError(f"Expected metadata's dtype is int32, got {md.dtype}")
    ##########################################################################
    # Verify stride.
    ##########################################################################
    nz_bitwidth = nz.dtype.itemsize * 8
    nz_packing = 32 // nz_bitwidth
    if sparse_dim:
        stride_multiple = 128
    else:
        stride_multiple = 8 * nz_packing
    if stride % stride_multiple != 0:
        raise ValueError(
            f"Expected {stride=} is a multiple of {stride_multiple}")
    ##########################################################################
    # Verify metadata compression produces expected shape.
    ##########################################################################
    name = "rhs" if rhs_sparse else "lhs"
    nz_x, nz_h, nz_w = nz.shape
    md_x, md_h, md_w = md.shape
    # For convenience (to avoid dealing with paddings), we try to select power of
    # 2 as nonzeross idx bitwidth. For example, if sparsity is 2:5, we will choose
    # 4 bits instead of 3.
    md_packing = _get_metadata_packing(y)
    if x != nz_x:
        raise ValueError(
            f"Expected {name} nonzeros shape[0] is {x}, got {nz_x}")
    if x != md_x:
        raise ValueError(
            f"Expected {name} metadata shape[0] is {x}, got {md_x}")
    if nz_w != md_w:
        raise ValueError(
            f"Expected nonzeros and metadata of {name} have same shape[2], got"
            f" {nz_w} vs {md_w}")
    if nz_h % md_packing != 0:
        raise ValueError(
            f"Expected {name} rows is a multiple of {md_packing=}, got {nz_h}")
    if nz_h // md_packing != md_h:
        raise ValueError(
            f"Expected {name} metadata rows can be perfectly packed by {md_packing}"
        )
    ##########################################################################
    # Verify nonzeros shape and matrix shape are expected.
    ##########################################################################
    decompressed_shape = list(nz.shape[1:])
    decompressed_shape[sparse_dim] *= y
    s_dims, d_dims = _get_dim_mapping(rhs_sparse, rhs_transpose)
    mnk = [None, None, None]
    for i in range(2):
        mnk[s_dims[i]] = decompressed_shape[i]
    for i in range(2):
        if mnk[d_dims[i]] and mnk[d_dims[i]] != mat.shape[i]:
            lhs_k = mnk[d_dims[i]]
            rhs_k = mat.shape[i]
            if rhs_transpose:
                lhs_k, rhs_k = rhs_k, lhs_k
            raise ValueError(
                f"Expected lhs contracting size {lhs_k} is equal to rhs contracting"
                f" size {rhs_k}")
        mnk[d_dims[i]] = mat.shape[i]
    m, n, k = mnk
    ##########################################################################
    # Verify block dimensions.
    ##########################################################################
    dim_names = ["m", "n", "k"]
    block_sizes = [block_m, block_n, block_k]
    block_multiples = [[], [], []]
    block_multiples[s_dims[sparse_dim]].append(stride * y)
    block_multiples[s_dims[0]].append(md_packing)

    for name, sz, bsz, multiples in zip(dim_names, mnk, block_sizes,
                                        block_multiples):
        for factor in multiples:
            if bsz % factor != 0:
                raise ValueError(
                    f"Expected block_{name}={bsz} is a multiple of {factor}")
        if sz % bsz != 0:
            raise ValueError(
                f"Expected block_{name}={bsz} can fully divide output rows {sz}"
            )
    return m, n, k


@functools.partial(
    jax.jit,
    static_argnames=[
        "sparsity",
        "sparse_dim",
        "rhs_sparse",
        "rhs_transpose",
        "stride",
        "block_m",
        "block_k",
        "block_n",
        "default_value",
        "out_dtype",
        "debug",
    ],
)
def _structured_spmm(
    sparsity: tuple[int, int],  # specify the x:y structured sparsity.
    nonzeros: jax.Array,
    metadata: jax.Array,
    matrix: jax.Array,
    *,
    sparse_dim: int,
    rhs_sparse: bool,
    rhs_transpose: bool = False,
    stride: int,
    block_m: int,
    block_k: int,
    block_n: int,
    default_value: Any = 0,
    out_dtype: Any = None,
    debug: bool = False,
) -> jax.Array:
    """General structured sparse matmul.

  Args:
    sparsity: the N:M structured sparsity.
    nonzeros: the nonzeros matrix.
    metadata: packed nonzeros idices.
    matrix: the dense matrix to matmul with.
    sparse_dim: sparsity is on each row when 0 or each column when 1.
    rhs_sparse: if true, nonzeros and metadata are from RHS and `matrix` is LHS.
    rhs_transpose: if true, RHS is transposed.
    stride: each value in N:M structured are separated in stride.
    block_m: the block size of LHS non-contracting dim.
    block_k: the block size of contracting dim.
    block_n: the block size of RHS non-contracting dim.
    default_value: the default value for the masked out data.
    out_dtype: the dtype of the output.
    debug: if true, enable debug mode in pallas kernel.

  Returns:
    The result of sparse matmul.

  Raises:
    ValueError: An error occurred passing invalid input.
  """
    m, n, k = _verify(
        sparsity,
        nonzeros,
        metadata,
        matrix,
        sparse_dim=sparse_dim,
        rhs_sparse=rhs_sparse,
        rhs_transpose=rhs_transpose,
        stride=stride,
        block_m=block_m,
        block_k=block_k,
        block_n=block_n,
    )

    _, y = sparsity
    md_packing = _get_metadata_packing(y)

    if not out_dtype:
        out_dtype = _infer_out_dtype(matrix.dtype, nonzeros.dtype)

    acc_dtype = jnp.float32
    if jnp.issubdtype(out_dtype, jnp.integer):
        acc_dtype = jnp.int32

    def _kernel(
        nz_tile_ref,
        md_tile_ref,
        mat_tile_ref,
        out_tile_ref,
        acc_tile_ref,
    ):

        @pl.when(pl.program_id(2) == 0)
        def _():
            acc_tile_ref[...] = jnp.zeros_like(acc_tile_ref, acc_dtype)

        # Decompress the metadata which contains nonzeros' indices.
        nonzeros_idx = _decompress_metadata(md_tile_ref, md_packing)
        # TODO(jevinjiang): try with large tiling.
        nonzeros = nz_tile_ref[...].astype(acc_dtype)
        decompressed_tile = _decompress_nonzeros(
            sparsity,
            nonzeros,
            nonzeros_idx,
            sparse_dim,
            stride,
            default_value,
        )
        mat_tile = mat_tile_ref[...]
        if rhs_sparse:
            lhs_tile, rhs_tile = mat_tile, decompressed_tile
        else:
            lhs_tile, rhs_tile = decompressed_tile, mat_tile
        # TODO(b/332970531): support arbitrary shape matmul for packed type.
        packing = 4 // lhs_tile.dtype.itemsize
        if packing > 1 and lhs_tile.shape[0] % (8 * packing) != 0:
            lhs_tile = lhs_tile.astype(acc_dtype)
        if packing > 1 and rhs_tile.shape[0] % (8 * packing) != 0:
            rhs_tile = rhs_tile.astype(acc_dtype)

        dim_nums = _get_dot_general_dim_nums(rhs_transpose)
        acc_tile_ref[...] += jax.lax.dot_general(
            lhs_tile, rhs_tile, dim_nums, preferred_element_type=acc_dtype)

        @pl.when(pl.program_id(2) == k // block_k - 1)
        def _():
            out_tile_ref[...] = acc_tile_ref[...].astype(out_dtype)

    in_specs = _get_in_sepcs(
        sparsity,
        sparse_dim,
        rhs_sparse,
        rhs_transpose,
        block_m,
        block_k,
        block_n,
    )
    return pl.pallas_call(
        _kernel,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=0,
            grid=(m // block_m, n // block_n, k // block_k),
            in_specs=in_specs,
            out_specs=pl.BlockSpec((block_m, block_n), lambda i, j, k: (i, j)),
            scratch_shapes=[pltpu.VMEM((block_m, block_n), acc_dtype)],
        ),
        out_shape=jax.ShapeDtypeStruct((m, n), out_dtype),
        debug=debug,
    )(nonzeros, metadata, matrix)


# TODO(jevinjiang): Deprecate this API.
@functools.partial(
    jax.jit,
    static_argnames=[
        "sparsity",
        "rhs_sparse",
        "contract_sparse",
        "rhs_transpose",
        "stride",
        "block_m",
        "block_k",
        "block_n",
        "default_value",
        "out_dtype",
        "debug",
    ],
)
def structured_spmm(
    sparsity: tuple[int, int],  # specify the x:y structured sparsity.
    nonzeros: jax.Array,
    metadata: jax.Array,
    matrix: jax.Array,
    *,
    rhs_sparse: bool,
    contract_sparse: bool,
    rhs_transpose: bool = False,
    stride: int,
    block_m: int,
    block_k: int,
    block_n: int,
    default_value: Any = 0,
    out_dtype: Any = None,
    debug: bool = False,
) -> jax.Array:
    """General structured sparse matmul.

  Args:
    sparsity: the N:M structured sparsity.
    nonzeros: the nonzeros matrix.
    metadata: packed nonzeros idices.
    matrix: the dense matrix to matmul with.
    rhs_sparse: if true, nonzeros and metadata are from RHS and `matrix` is LHS.
    contract_sparse: if true, the values are sparse on the contracting dimension
    rhs_transpose: if true, RHS is transposed.
    stride: each value in N:M structured are separated in stride.
    block_m: the block size of LHS non-contracting dim.
    block_k: the block size of contracting dim.
    block_n: the block size of RHS non-contracting dim.
    default_value: the default value for the masked out data.
    out_dtype: the dtype of the output.
    debug: if true, enable debug mode in pallas kernel.

  Returns:
    The result of sparse matmul.

  Raises:
    ValueError: An error occurred passing invalid input.
  """
    # TODO(jevinjiang): Deprecate contract_sparse to use sparse_dim directly.
    if rhs_sparse:
        if contract_sparse:
            sparse_dim = 1 if rhs_transpose else 0
        else:
            sparse_dim = 0 if rhs_transpose else 1
    else:
        if contract_sparse:
            sparse_dim = 1
        else:
            sparse_dim = 0
    return _structured_spmm(
        sparsity,
        nonzeros,
        metadata,
        matrix,
        sparse_dim=sparse_dim,
        rhs_sparse=rhs_sparse,
        rhs_transpose=rhs_transpose,
        stride=stride,
        block_m=block_m,
        block_k=block_k,
        block_n=block_n,
        default_value=default_value,
        out_dtype=out_dtype,
        debug=debug,
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.experimental.pallas as pl
import jax.experimental.pallas.tpu as pltpu
import jax.lax as lax
import jax.numpy as jnp
from jax._src.pallas.mosaic import pipeline

from tpu_inference.kernels.experimental.batched_rpa import (bref_override,
                                                            flash_attention)
from tpu_inference.kernels.experimental.batched_rpa import \
    schedule as schedule_lib
from tpu_inference.kernels.experimental.batched_rpa import utils


def make_rpa_kernel(config: schedule_lib.RPAConfig):
    q_packing = schedule_lib.get_dtype_packing(config.q_dtype)
    kv_packing = schedule_lib.get_dtype_packing(config.kv_dtype)

    def ragged_paged_attention_pipeline(
        # metadata inputs
        cu_q_lens_ref,
        kv_lens_ref,
        page_indices_ref,
        # schedule flattened (9 arrays)
        schedule_hbm,
        # hbm inputs
        o_hbm_alias_q_hbm_ref,
        new_kv_hbm_ref,
        kv_cache_hbm_ref,
        # output
        o_hbm_ref,
        kv_cache_out_ref,
        # scratch allocations
        schedule,
        m_scratch,
        l_scratch,
        acc_scratch,
        dma_sem,
    ):

        # q_hbm_ref shape: [q_times_kv, max_tokens, d] or similar
        # We use config values for dimensions where possible, or infer from refs
        kv_heads = config.num_kv_heads
        q_per_kv = config.num_q_heads_per_kv_head
        d_dim = config.head_dim
        _, _, num_kv_x2_packed, _, _ = kv_cache_hbm_ref.shape

        q_vmem_shape = (
            config.batch_size,
            kv_heads,
            config.bq_sz,
            q_per_kv // q_packing,
            q_packing,
            d_dim,
        )
        bkv_stride: int = (kv_heads * 2) // kv_packing

        if schedule_lib.has_bank_conflicts(bkv_stride):
            bkv_stride += 1

        kv_vmem_shape = (
            config.batch_size,
            config.bkv_sz,
            bkv_stride,
            kv_packing,
            d_dim,
        )

        def rpa_body(
            # input
            q_vref,
            kv_in_vref,
            # output
            o_vref,
        ):

            def _strided_load_bkv(b_idx, start):
                assert start % kv_packing == 0
                start //= kv_packing
                kv_ref = (kv_in_vref.bitcast(jnp.uint32).at[b_idx].reshape(
                    config.bkv_sz * bkv_stride, config.head_dim))

                if kv_packing == 1:
                    k = utils.strided_load(
                        kv_ref,
                        start,
                        config.bkv_sz * bkv_stride,
                        bkv_stride,
                        dtype=config.kv_dtype,
                    )
                    v = utils.strided_load(
                        kv_ref,
                        start + 1,
                        config.bkv_sz * bkv_stride,
                        bkv_stride,
                        dtype=config.kv_dtype,
                    )
                    return [(k, v)]

                kv = utils.strided_load(kv_ref, start,
                                        config.bkv_sz * bkv_stride, bkv_stride)
                bitwidth = 32 // kv_packing

                return utils.convert_to_target_bitwidth(
                    kv, target_bitwidth=bitwidth, kv_dtype=config.kv_dtype)

            step = pl.program_id(0)

            processed_q_len = []
            processed_kv_len = []
            effective_kv_len = []
            int_ty = config.int_ty
            for b in range(config.batch_size):
                idx = step * config.batch_size + b
                s_idx = schedule.s_idx[idx]
                is_valid = s_idx != -1
                q_idx = schedule.q_idx[idx]
                k_idx = schedule.k_idx[idx]
                k_id = lax.select(is_valid, k_idx * config.bkv_sz, 0)
                kv_len = lax.select(is_valid, kv_lens_ref[s_idx], 0)
                q_start = lax.select(is_valid, cu_q_lens_ref[s_idx], 0)
                q_end = lax.select(is_valid, cu_q_lens_ref[s_idx + 1], 0)
                q_len = q_end - q_start
                offset = kv_len - q_len

                processed_q_len.append(
                    (q_idx * config.bq_sz + offset).astype(int_ty))
                processed_kv_len.append(k_id.astype(int_ty))
                effective_kv_len.append(kv_len.astype(int_ty))

                start_k_idx = 0
                if config.sliding_window is not None:
                    sw_start_idx = (kv_len - q_len + q_idx * config.bq_sz -
                                    config.sliding_window + 1)
                    start_k_idx = jnp.maximum(0, sw_start_idx) // config.bkv_sz

                is_first_k_block = k_idx == start_k_idx
                reset_cond = is_valid & is_first_k_block
                m_scratch[b] = jnp.where(reset_cond, -jnp.inf, m_scratch[b])
                l_scratch[b] = jnp.where(reset_cond, 0.0, l_scratch[b])
                acc_scratch[b] = jnp.where(reset_cond, 0.0, acc_scratch[b])

            q_p = config.num_q_heads_per_kv_head // q_packing
            q_ref = q_vref.bitcast(jnp.uint32).reshape(-1, config.head_dim)
            q_loaded = utils.strided_load(
                q_ref,
                0,
                config.batch_size * config.num_kv_heads * config.bq_sz * q_p,
                1,
                dtype=config.q_dtype,
            )
            q = q_loaded.reshape(
                config.batch_size,
                config.num_kv_heads,
                config.bq_sz * config.num_q_heads_per_kv_head,
                config.head_dim,
            )

            # We want to load k, v from (batch, bkv_sz, bkv_stride, kv_packing, d)
            # where bkv_stride ~= num_kv_heads * 2 // kv_packing
            # to 2x (batch, num_kv_heads, bkv_sz, d)
            # We use strided_load to avoid the expensive transpose.
            k_b = []
            v_b = []
            for b in range(config.batch_size):
                heads_per_load = max(1, kv_packing // 2)
                ks = []
                vs = []
                for kv_head_start in range(0, config.num_kv_heads,
                                           heads_per_load):
                    bkv_lst = _strided_load_bkv(
                        b,
                        kv_head_start * 2,
                    )
                    ks.append(jnp.stack([k for k, _ in bkv_lst], axis=0))
                    vs.append(jnp.stack([v for _, v in bkv_lst], axis=0))
                k, v = jnp.concatenate(ks, axis=0), jnp.concatenate(vs, axis=0)
                k = k.reshape(-1, config.bkv_sz, config.head_dim)
                v = v.reshape(-1, config.bkv_sz, config.head_dim)

                k = k[:config.num_kv_heads]
                v = v[:config.num_kv_heads]
                k_b.append(k)
                v_b.append(v)
            # Stack to (batch, num_heads, bkv_sz, 128)
            k = jnp.stack(k_b, axis=0)
            v = jnp.stack(v_b, axis=0)

            m_val = m_scratch[...]
            l_val = l_scratch[...]
            acc_val = acc_scratch[...]

            m_next, l_next, o_next = flash_attention.flash_attention(
                q,
                k,
                v,
                acc_val,
                m_val,
                l_val,
                processed_q_len=processed_q_len,
                processed_kv_len=processed_kv_len,
                effective_kv_len=effective_kv_len,
                config=config,
            )
            m_scratch[...] = m_next
            l_scratch[...] = l_next
            acc_scratch[...] = o_next

            @pl.loop(0, config.batch_size, unroll=True)
            def _for_each_row(b):

                def _accum():
                    o = acc_scratch[b]
                    l_ = l_scratch[b]
                    l_ = jnp.tile(l_, (o.shape[-1] // l_.shape[-1], ))
                    if config.out_dtype == jnp.float32:
                        result = lax.div(o, l_)
                    else:
                        result = (o * pl.reciprocal(l_, approx=True) if
                                  (l_.dtype == jnp.float32
                                   and config.out_dtype != jnp.float32) else
                                  lax.div(o, l_)).astype(config.out_dtype)

                    out = result.astype(o_vref.dtype)
                    out_ref = (o_vref.at[b].bitcast(jnp.int32).reshape(
                        config.num_kv_heads * config.bq_sz *
                        (config.num_q_heads_per_kv_head // q_packing),
                        config.head_dim,
                    ))
                    out = pltpu.bitcast(out,
                                        out_ref.dtype).reshape(out_ref.shape)
                    utils.strided_store(out_ref, 0, out_ref.shape[0], 1, out)

                # Adding a conditional causes a scheduling barrier. In prefill, we often
                # use small block sizes, so it's not worth executing the accumulation
                # on every block. In decode, because of the large block sizes / and or
                # batch sizes, we almost always use accumulation on every block. Please
                # tune `fuse_accum` for your use case.
                if not config.fuse_accum:
                    idx = step * config.batch_size + b
                    is_last_k = schedule.is_last_k[idx] == 1
                    jax.lax.cond(is_last_k, _accum, lambda: None)
                else:
                    _accum()

        kv_cache_spec = pl.BlockSpec(
            block_shape=kv_vmem_shape,
            memory_space=pltpu.VMEM,
            index_map=lambda i: (i, ),
            pipeline_mode=pl.Buffered(buffer_count=config.n_buffer,
                                      use_lookahead=True),
        )
        q_spec = pl.BlockSpec(
            block_shape=q_vmem_shape,
            memory_space=pltpu.VMEM,
            index_map=lambda i: (i, ),
            pipeline_mode=pl.Buffered(buffer_count=config.n_buffer,
                                      use_lookahead=True),
        )
        o_spec = pl.BlockSpec(
            block_shape=q_vmem_shape,
            memory_space=pltpu.VMEM,
            index_map=lambda i: (i, ),
            pipeline_mode=pl.Buffered(buffer_count=2, use_lookahead=False),
        )

        # hbm_kv_packed_stride = (config.num_kv_heads * 2 + kv_packing - 1) // kv_packing
        kv_cache_alloc = bref_override.KVBufferedRef.create(
            spec=kv_cache_spec,
            source_memory_space=kv_cache_hbm_ref,
            bkv_p_cache=config.bkv_p_cache,
            bkv_p_new=config.bkv_p_new,
            page_size=config.page_size,
            batch_size=config.batch_size,
            hbm_stride=num_kv_x2_packed,
            page_size_log2=config.page_size_log2,
            page_size_mask=config.page_size_mask,
            buffer_count=config.n_buffer,
            use_lookahead=True,
        )
        q_alloc = bref_override.BatchingQRef.create(
            spec=q_spec,
            source_memory_space=o_hbm_alias_q_hbm_ref,
            bq_sz=config.bq_sz,
            batch_size=config.batch_size,
            buffer_count=config.n_buffer,
            use_lookahead=True,
        )
        o_alloc = bref_override.BatchingORef.create(
            spec=o_spec,
            source_memory_space=o_hbm_ref,
            batch_size=config.batch_size,
            buffer_count=2,
            use_lookahead=False,
        )

        def _run(final_allocs):
            actual_steps = schedule_hbm.actual_steps[0]
            safe_steps = jnp.minimum(actual_steps, config.max_steps_ub)
            grid = (safe_steps, )

            # Transfer schedule from HBM to SMEM --- we only copy what we need. Since
            # we almost always over-allocate schedule size, we only want to copy a small
            # portion of it from HBM to SMEM.
            sem = dma_sem.at[0]
            flat_hbm = jax.tree_util.tree_leaves(schedule_hbm)
            flat_smem = jax.tree_util.tree_leaves(schedule)
            dma_list = []
            for h, s in zip(flat_hbm, flat_smem):
                if h.shape[0] != 1:
                    fetch_size = (h.shape[0] //
                                  config.max_steps_ub) * safe_steps
                    fetch_size = pl.cdiv(fetch_size, 1024) * 1024

                    copy = pltpu.make_async_copy(
                        h.at[pl.ds(0, fetch_size)],
                        s.at[pl.ds(0, fetch_size)],
                        sem,
                    )
                    copy.start()
                    dma_list.append(copy)

            # Initialize KV cache to zeros
            kv_alloc = final_allocs[1]
            num_lanes = pltpu.get_tpu_info().num_lanes
            kv_ref_flat = kv_alloc.window_ref.bitcast(jnp.uint32).reshape(
                -1, num_lanes)
            kv_ref_flat[...] = jnp.zeros_like(kv_ref_flat)

            for copy in dma_list:
                copy.wait()

            pipeline_func = pipeline.emit_pipeline(
                body=rpa_body,
                grid=grid,
                in_specs=[
                    q_spec,
                    kv_cache_spec,
                ],
                out_specs=[
                    o_spec,
                ],
            )

            pipeline_func(
                (o_hbm_alias_q_hbm_ref, schedule),
                (kv_cache_hbm_ref, new_kv_hbm_ref, schedule, page_indices_ref),
                (o_hbm_ref, schedule),
                allocations=final_allocs,
            )

        return pl.run_scoped(_run, (q_alloc, kv_cache_alloc, o_alloc))

    num_pages = config.num_seq * config.pages_per_seq
    if config.total_num_pages is not None:
        num_pages = config.total_num_pages
    num_kv_heads_x2_packed = (config.num_kv_heads * 2 + kv_packing -
                              1) // kv_packing

    out_shape = [
        jax.ShapeDtypeStruct(
            (
                config.num_kv_heads,
                config.total_q_tokens,
                config.num_q_heads_per_kv_head // q_packing,
                q_packing,
                config.head_dim,
            ),
            config.q_dtype,
        ),
        jax.ShapeDtypeStruct(
            (
                num_pages,
                config.page_size,
                num_kv_heads_x2_packed,
                kv_packing,
                config.head_dim,
            ),
            config.kv_dtype,
        ),
    ]

    schedule_shapes = schedule_lib.RPASchedule.smem_specs(config)
    scratch_shapes = [
        schedule_shapes,
        pltpu.VMEM(
            (
                config.batch_size,
                config.num_kv_heads,
                config.bq_sz * config.num_q_heads_per_kv_head,
                128,
            ),
            dtype=config.out_dtype,
        ),  # m
        pltpu.VMEM(
            (
                config.batch_size,
                config.num_kv_heads,
                config.bq_sz * config.num_q_heads_per_kv_head,
                128,
            ),
            dtype=config.out_dtype,
        ),  # l
        pltpu.VMEM(
            (
                config.batch_size,
                config.num_kv_heads,
                config.bq_sz * config.num_q_heads_per_kv_head,
                config.head_dim,
            ),
            dtype=config.out_dtype,
        ),  # acc
        pltpu.SemaphoreType.DMA((1, )),  # dma_sem
    ]
    in_specs = [
        schedule_lib.RPASchedule.kernel_in_specs(config),  # 9 refs
        pl.BlockSpec(memory_space=pltpu.HBM),  # o_hbm_alias_q_hbm_ref
        pl.BlockSpec(memory_space=pltpu.HBM),  # new_kv_hbm_ref
        pl.BlockSpec(memory_space=pltpu.HBM),  # kv_cache_hbm_ref
    ]
    input_output_aliases = {12: 0, 14: 1}

    scope_name = f"RPA{config.case.symbol}-p{config.page_size}-b{config.batch_size}-q{config.bq_sz}-k{config.bkv_sz}"
    if config.sliding_window:
        scope_name += f"-sw{config.sliding_window}"
    _kernel = pl.pallas_call(
        ragged_paged_attention_pipeline,
        out_shape=out_shape,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            grid=(1, ),
            num_scalar_prefetch=3,
            in_specs=in_specs,
            out_specs=[
                pl.BlockSpec(memory_space=pltpu.HBM),  # o_hbm_ref
                pl.BlockSpec(memory_space=pltpu.HBM),  # kv_cache_out_ref
            ],
            scratch_shapes=scratch_shapes,
        ),
        compiler_params=pltpu.CompilerParams(
            dimension_semantics=("arbitrary", ),
            vmem_limit_bytes=config.vmem_limit_bytes,
            disable_bounds_checks=True,
        ),
        input_output_aliases=input_output_aliases,
        name=scope_name,
    )

    def _wrap_kernel_in_hbm_constraints(
        cu_q_lens_ref,
        kv_lens_ref,
        page_indices_ref,
        schedule_hbm_ref,
        o_hbm_alias_q_hbm_ref,
        new_kv_hbm_ref,
        kv_cache_hbm_ref,
    ):

        def _constrain_hbm(path, x):
            for p in path:
                key = getattr(p, "name", getattr(p, "key", None))
                if key == "actual_steps":
                    return x
            return pltpu.with_memory_space_constraint(x, pltpu.HBM)

        constrained_schedule_hbm_ref = jax.tree_util.tree_map_with_path(
            _constrain_hbm, schedule_hbm_ref)
        return _kernel(
            cu_q_lens_ref,
            kv_lens_ref,
            page_indices_ref,
            constrained_schedule_hbm_ref,
            pltpu.with_memory_space_constraint(o_hbm_alias_q_hbm_ref,
                                               pltpu.HBM),
            pltpu.with_memory_space_constraint(new_kv_hbm_ref, pltpu.HBM),
            pltpu.with_memory_space_constraint(kv_cache_hbm_ref, pltpu.HBM),
        )

    return _wrap_kernel_in_hbm_constraints



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.experimental.pallas as pl
import jax.experimental.pallas.tpu as pltpu
import jax.numpy as jnp

from tpu_inference.kernels.experimental.batched_rpa import \
    schedule as schedule_lib


def align_to(a, b):
    """Returns 'a' aligned to 'b'."""
    return pl.cdiv(a, b) * b


def broadcast_minor(src, shape):
    """Broadcasts 'src' to 'shape' in the minor dimension."""
    if src.shape == shape:
        return src
    assert src.shape[:-1] == shape[:-1]
    assert src.shape[-1] % 128 == 0
    target_minor = align_to(shape[-1], src.shape[-1])
    # no-op concatenation.
    return jnp.concatenate([src for _ in range(target_minor // src.shape[-1])],
                           axis=-1)[..., :shape[-1]]


def get_dtype_bitwidth(dtype):
    """Returns the bitwidth of a JAX dtype."""
    return jax._src.dtypes.itemsize_bits(dtype)


def strided_load(ref, start_row, num_rows, step, *, dtype=None):
    """Loads data from HBM with strided access, handling 128-lane alignment."""
    _, row_width = ref.shape
    num_sub_lanes = row_width // 128
    ref_flat = ref.reshape(-1, 128)

    # scale indices to match flattened arraw.
    v_start = start_row * num_sub_lanes
    v_num = num_rows * num_sub_lanes
    v_step = step * num_sub_lanes

    # Gather the chunks into the original head dimension.
    chunks = [
        ref_flat[pl.ds(v_start + i, v_num // v_step, v_step)]
        for i in range(num_sub_lanes)
    ]
    vec = jnp.concat(chunks, axis=1)

    return pltpu.bitcast(vec, dtype) if dtype is not None else vec


def strided_store(ref, start, sz, step, val):
    """Stores data to HBM with strided access, handling 128-lane alignment."""
    assert schedule_lib.get_dtype_packing(ref.dtype) == 1
    assert ref.dtype == val.dtype
    assert ref.shape == val.shape
    assert len(ref.shape) == 2
    r, l = ref.shape  # noqa
    assert l % 128 == 0
    folds = l // 128
    ref = ref.reshape(r * folds, 128)
    start *= folds
    sz *= folds
    step *= folds
    assert sz % step == 0
    for i in range(folds):
        ref[pl.ds(start + i, sz // step, step)] = val[:, i * 128:(i + 1) * 128]


# If we want to convert 32-bits into 32//N number of N-bits value, naive
# approach would be to perform 32//N number of 32-bits to N-bits conversion.
# However, we can reduce number of instructions by utilizing binary tree.
# 0: [32]
# 1: [16, 16]
# ...
# log2(32//N): [N, N, ... N]
def convert_to_target_bitwidth(val, target_bitwidth: int, kv_dtype: jnp.dtype):
    """Converts a value to a target bitwidth."""
    curr_dtype = val.dtype
    curr_bitwidth = get_dtype_bitwidth(curr_dtype)
    assert target_bitwidth != curr_bitwidth, "No conversion is needed."

    # We split val into two vals (left and right) where each have half of the
    # original bitwidth.
    next_bitwidth = curr_bitwidth // 2
    next_dtype = jnp.dtype(f"uint{next_bitwidth}")

    left = val.astype(next_dtype)

    # Bitwise shift is only supported in uint32.
    val_u32 = pltpu.bitcast(val, jnp.uint32)
    val_u32_shifted = val_u32 >> next_bitwidth
    # Convert back to original dtype.
    val_shifted = pltpu.bitcast(val_u32_shifted, curr_dtype)
    right = val_shifted.astype(next_dtype)

    if next_bitwidth == target_bitwidth:
        k = pltpu.bitcast(left, kv_dtype)
        v = pltpu.bitcast(right, kv_dtype)
        return [(k, v)]
    else:
        left_out = convert_to_target_bitwidth(left,
                                              target_bitwidth=target_bitwidth,
                                              kv_dtype=kv_dtype)
        right_out = convert_to_target_bitwidth(right,
                                               target_bitwidth=target_bitwidth,
                                               kv_dtype=kv_dtype)
        return left_out + right_out



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P
from transformers import Qwen3Config
from vllm.config import VllmConfig

from tpu_inference.distributed.jax_parallel_state import get_pp_group
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.embed import JaxEmbed
from tpu_inference.layers.jax.linear import JaxEinsum, JaxLinear
from tpu_inference.layers.jax.moe.moe import JaxMoE
from tpu_inference.layers.jax.moe.utils import (get_expert_parallelism,
                                                select_moe_backend)
from tpu_inference.layers.jax.norm import JaxRmsNorm
from tpu_inference.layers.jax.pp_utils import PPMissingLayer, make_layers
from tpu_inference.layers.jax.quantization.configs import QuantizationConfig
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
from tpu_inference.models.jax.qwen3 import Qwen3Attention
from tpu_inference.models.jax.utils.weight_utils import LoadableWithIterator

logger = init_logger(__name__)

init_fn = nnx.initializers.uniform()


class Qwen3MoeSparseMoeBlock(JaxModule):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 prefix: str = ""):
        config = vllm_config.model_config.hf_text_config
        dtype = vllm_config.model_config.dtype
        quant_config = vllm_config.quant_config

        # --- Sharding Config ---
        edf_sharding = (None, None, None)
        expert_axis_name = edf_sharding[0]
        num_expert_parallelism = get_expert_parallelism(expert_axis_name, mesh)
        use_ep = num_expert_parallelism > 1
        moe_backend = select_moe_backend(use_ep)

        # Router
        self.gate = JaxLinear(
            config.hidden_size,
            config.num_experts,
            dtype=dtype,
            param_dtype=dtype,
            rngs=rng,
            use_bias=False,
            quant_config=quant_config,
            prefix=prefix + ".gate",
        )
        self.gate.num_experts_per_tok = config.num_experts_per_tok

        # Shared Expert
        shared_expert_intermediate_size = getattr(
            config, "shared_expert_intermediate_size", 0)
        if shared_expert_intermediate_size > 0:
            raise NotImplementedError(
                f"Shared expert is not implemented yet. Found {shared_expert_intermediate_size=} in config."
            )
        else:
            self.shared_expert = None

        # Experts (Routed)
        self.experts = JaxMoE(
            dtype=dtype,
            num_local_experts=config.num_experts,
            hidden_size=config.hidden_size,
            intermediate_size_moe=config.moe_intermediate_size,
            hidden_act=config.hidden_act,
            rngs=rng,
            router=self.gate,
            mesh=mesh,
            activation_ffw_td=P(ShardingAxisName.MLP_DATA, None),
            activation_ffw_ted=P(ShardingAxisName.MLP_DATA, None, None),
            edf_sharding=P(None, ),
            efd_sharding=P(None, ),
            apply_expert_weight_before_computation=False,
            expert_axis_name=expert_axis_name,
            num_expert_parallelism=num_expert_parallelism,
            moe_backend=moe_backend,
            quant_config=quant_config,
            prefix=prefix + ".experts")

    def __call__(self, x: jax.Array) -> jax.Array:
        out = self.experts(x)
        if self.shared_expert is not None:
            out += self.shared_expert(x)
        return out


class Qwen3MoeDecoderLayer(JaxModule):

    def __init__(self,
                 config: Qwen3Config,
                 dtype: jnp.dtype,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 kv_cache_dtype: str,
                 quant_config: QuantizationConfig,
                 layer_idx: int,
                 vllm_config: VllmConfig,
                 prefix: str = ""):
        rms_norm_eps = config.rms_norm_eps
        hidden_size = config.hidden_size

        self.input_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            dtype=dtype,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".input_layernorm",
        )
        self.self_attn = Qwen3Attention(
            config=config,
            dtype=dtype,
            rng=rng,
            mesh=mesh,
            kv_cache_dtype=kv_cache_dtype,
            quant_config=quant_config,
            prefix=prefix + ".self_attn",
        )
        self.post_attention_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            param_dtype=dtype,
            dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".post_attention_layernorm",
        )

        mlp_only_layers = getattr(config, "mlp_only_layers", [])
        if (layer_idx not in mlp_only_layers) and (
                config.num_experts > 0 and
            (layer_idx + 1) % config.decoder_sparse_step == 0):
            self.mlp = Qwen3MoeSparseMoeBlock(vllm_config=vllm_config,
                                              rng=rng,
                                              mesh=mesh,
                                              prefix=prefix + ".mlp")
        else:
            raise NotImplementedError(
                f"Non-sparse MLP is not implemented yet. Found {mlp_only_layers=}, {config.num_experts=}, and {config.decoder_sparse_step=} in config."
            )

    def __call__(
        self,
        kv_cache: jax.Array,
        x: jax.Array,
        attention_metadata: AttentionMetadata,
    ) -> Tuple[jax.Array, jax.Array]:
        hidden_states = self.input_layernorm(x)
        kv_cache, attn_output = self.self_attn(
            kv_cache,
            hidden_states,
            attention_metadata,
        )
        attn_output += x

        residual = attn_output
        attn_output = self.post_attention_layernorm(attn_output)
        outputs = self.mlp(attn_output)
        outputs = residual + outputs
        return kv_cache, outputs


class Qwen3MoeModel(JaxModule):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 prefix: str = "") -> None:
        model_config = vllm_config.model_config
        hf_config = model_config.hf_config
        vocab_size = model_config.get_vocab_size()
        dtype = model_config.dtype
        rms_norm_eps = hf_config.rms_norm_eps
        hidden_size = hf_config.hidden_size

        self.is_first_rank = get_pp_group().is_first_rank
        self.is_last_rank = get_pp_group().is_last_rank

        if self.is_first_rank or (hf_config.tie_word_embeddings
                                  and self.is_last_rank):
            self.embed_tokens = JaxEmbed(
                num_embeddings=vocab_size,
                features=hidden_size,
                dtype=dtype,
                param_dtype=dtype,
                embedding_init=nnx.with_partitioning(init_fn, ("model", None)),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".embed_tokens",
            )
        else:
            self.embed_tokens = PPMissingLayer()

        self.start_layer, self.end_layer, self.layers = make_layers(
            hf_config.num_hidden_layers,
            lambda layer_index: Qwen3MoeDecoderLayer(
                config=hf_config,
                dtype=dtype,
                rng=rng,
                mesh=mesh,
                kv_cache_dtype=vllm_config.cache_config.cache_dtype,
                quant_config=vllm_config.quant_config,
                layer_idx=layer_index,
                vllm_config=vllm_config,
                prefix=f"{prefix}.layers.{layer_index}",
            ))

        if self.is_last_rank:
            self.norm = JaxRmsNorm(
                hidden_size,
                epsilon=rms_norm_eps,
                dtype=dtype,
                param_dtype=dtype,
                scale_init=nnx.with_partitioning(init_fn, (None, )),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".final_layernorm",
            )
        else:
            self.norm = PPMissingLayer()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
    ) -> Tuple[List[jax.Array], jax.Array]:
        if self.is_first_rank:
            assert inputs_embeds is None
            inputs_embeds = self.embed_tokens(input_ids)
        else:
            assert inputs_embeds is not None

        x = inputs_embeds
        new_kv_caches = []
        for i, layer in enumerate(self.layers):
            if isinstance(layer, PPMissingLayer):
                new_kv_caches.append(kv_caches[i])
                continue
            kv_cache = kv_caches[i]
            kv_cache, x = layer(kv_cache, x, attention_metadata)
            new_kv_caches.append(kv_cache)

        if self.is_last_rank:
            x = self.norm(x)

        return new_kv_caches, x


class Qwen3MoeForCausalLM(JaxModule, LoadableWithIterator):

    def __init__(self, vllm_config: VllmConfig, rng_key: jax.Array,
                 mesh: Mesh) -> None:
        if getattr(vllm_config.model_config, "quantization", None) == "fp8":
            # `get_tpu_quantization_config` returns None for "fp8" because
            # the work in #1623 is not fully merged. So this block overrides
            # the logic to return Fp8Config when model_config indicates fp8.
            # TODO(#1623): Remove this block when `get_tpu_quantization_config`
            # is updated.
            from tpu_inference.layers.jax.quantization.fp8 import Fp8Config
            hg_quant_config = getattr(vllm_config.model_config.hf_config,
                                      "quantization_config", {})
            vllm_config.quant_config = Fp8Config(hg_quant_config)

        self.vllm_config = vllm_config
        rng = nnx.Rngs(rng_key)
        self.mesh = mesh

        self.model = Qwen3MoeModel(
            vllm_config=vllm_config,
            rng=rng,
            mesh=mesh,
            prefix="model",
        )
        model_config = vllm_config.model_config
        if not model_config.hf_config.tie_word_embeddings:
            if self.model.is_last_rank:
                vocab_size = model_config.get_vocab_size()
                hidden_size = model_config.hf_config.hidden_size
                self.lm_head = JaxEinsum(
                    einsum_str="TD,DV->TV",
                    kernel_shape=(hidden_size, vocab_size),
                    dtype=model_config.dtype,
                    param_dtype=model_config.dtype,
                    rngs=rng,
                    quant_config=vllm_config.quant_config,
                    prefix="lm_head",
                )
            else:
                self.lm_head = PPMissingLayer()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
        _input_positions=None,
        _layer_name_to_kv_cache=None,
        _lora_metadata=None,
        intermediate_tensors: JaxIntermediateTensors | None = None,
        is_first_rank: bool = True,
        is_last_rank: bool = True,
        *args,
    ) -> Tuple[List[jax.Array], jax.Array | JaxIntermediateTensors,
               List[jax.Array]]:
        if not is_first_rank:
            assert intermediate_tensors is not None
            inputs_embeds = intermediate_tensors["hidden_states"]
        kv_caches, x = self.model(
            kv_caches,
            input_ids,
            attention_metadata,
            inputs_embeds,
        )
        if not is_last_rank:
            x = JaxIntermediateTensors(tensors={"hidden_states": x}, )
        return kv_caches, x, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        if hasattr(self, 'lm_head'):
            return self.lm_head(hidden_states)

        assert isinstance(self.model.embed_tokens, JaxEmbed)
        return self.model.embed_tokens.decode(hidden_states)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import jax
import jax.numpy as jnp
import numpy as np
from flax import nnx
from jax.sharding import Mesh

from tpu_inference.layers.jax.layers import DenseFFW, Embedder, RMSNorm


class TestLayers(unittest.TestCase):
    """Unit test suite for common JAX layer blocks."""

    def setUp(self):
        """Sets up the testing environment before each test."""
        self.mesh = Mesh(
            np.array(jax.devices()).reshape(1, -1),
            axis_names=(
                "expert",
                "model",
            ),
        )

    def test_rmsnorm_forward_pass(self):
        """Tests the forward pass of the RMSNorm module."""
        with jax.set_mesh(self.mesh):
            dims = 512
            epsilon = 1e-5

            norm = RMSNorm(
                dims=dims,
                random_init=True,
                epsilon=epsilon,
                rngs=nnx.Rngs(0),
                dtype=jnp.float32,
            )

            seq_len = 128
            x = jax.random.normal(jax.random.PRNGKey(42), (seq_len, dims))

            output = norm(x)

            self.assertEqual(output.shape, x.shape)
            self.assertEqual(output.dtype, jnp.float32)

            mean_of_squares = jnp.mean(jnp.square(output), axis=-1)
            self.assertTrue(
                jnp.allclose(mean_of_squares, 1.0, atol=1e-5).all())

    def test_denseffw_forward_pass(self):
        """Tests the forward pass of the DenseFFW module."""
        with jax.set_mesh(self.mesh):
            hidden_size = 512
            intermediate_size = 2048

            ffw_layer = DenseFFW(random_init=True,
                                 dtype=jnp.bfloat16,
                                 hidden_act="silu",
                                 hidden_size=hidden_size,
                                 intermediate_size=intermediate_size,
                                 rngs=nnx.Rngs(0),
                                 mesh=self.mesh)

            seq_len = 128
            x = jnp.ones((seq_len, hidden_size), dtype=jnp.bfloat16)

            output = ffw_layer(x)

            self.assertEqual(output.shape, x.shape)
            self.assertEqual(output.dtype, x.dtype)

    def test_embedder_forward_pass(self):
        """Tests both the encode and decode passes of the Embedder module."""
        with jax.set_mesh(self.mesh):
            hidden_size = 512
            vocab_size = 32000
            dtype = jnp.bfloat16

            embedder = Embedder(
                vocab_size=vocab_size,
                hidden_size=hidden_size,
                dtype=dtype,
                random_init=True,
                rngs=nnx.Rngs(0),
            )

            seq_len = 128
            token_ids = jnp.arange(seq_len, dtype=jnp.int32) % vocab_size
            embeddings = embedder(token_ids, decode=False)
            self.assertEqual(embeddings.shape, (seq_len, hidden_size))
            self.assertEqual(embeddings.dtype, dtype)

            hidden_states = jnp.ones((seq_len, hidden_size),
                                     dtype=jnp.bfloat16)
            logits = embedder(hidden_states, decode=True)
            self.assertEqual(logits.shape, (seq_len, vocab_size))
            self.assertEqual(logits.dtype, dtype)

    def test_embedder_normalization(self):
        """Tests the embedding normalization feature."""
        with jax.set_mesh(self.mesh):
            hidden_size = 512
            vocab_size = 32000

            rngs_1 = nnx.Rngs(42)
            rngs_2 = nnx.Rngs(42)

            embedder_norm = Embedder(
                vocab_size=vocab_size,
                hidden_size=hidden_size,
                dtype=jnp.float32,
                normalize_embeddings=True,
                random_init=True,
                rngs=rngs_1,
            )

            embedder_no_norm = Embedder(
                vocab_size=vocab_size,
                hidden_size=hidden_size,
                dtype=jnp.float32,
                random_init=True,
                rngs=rngs_2,
            )

            token_ids = jnp.arange(10, dtype=jnp.int32)

            embeddings_norm = embedder_norm(token_ids, decode=False)
            embeddings_no_norm = embedder_no_norm(token_ids, decode=False)

            scaling_factor = jnp.sqrt(hidden_size)
            expected_embeddings = embeddings_no_norm * scaling_factor

            self.assertTrue(
                jnp.allclose(embeddings_norm, expected_embeddings,
                             atol=1e-6).all())


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import Mesh, PartitionSpec
from qwix._src.providers import ptq
from vllm.model_executor.layers.fused_moe import FusedMoE

from tpu_inference.layers.common.process_weights.moe_weights import \
    UnfusedMoEWeights
# yapf: disable
from tpu_inference.layers.jax.moe.utils import (get_all_to_all_params_fn,
                                                global_permute_fn, gmm_fn,
                                                local_permute_fn,
                                                modeling_flax_utils,
                                                sort_activations_fn,
                                                unpermute_fn)
from tpu_inference.models.jax.utils.qwix.qwix_utils import \
    manually_quantize_qwix_weight

if TYPE_CHECKING:
    from tpu_inference.layers.jax.moe.moe import JaxMoE
else:
    JaxMoE = None
# yapf: enable


def sparse_moe_distributed_fwd(
    moe_instance,
    x_TD: jax.Array,
    router_weights_TX: jax.Array,
    selected_experts_TX: jax.Array,
    kernel_gating: jax.Array,
    kernel_up_proj: jax.Array,
    kernel_down_proj: jax.Array,
):
    """
    The sparse MoE forward pass with fully distributed logic.
    This assumes it is running within a distributed TPU.
    """

    # 1. Global Permute
    (
        sorted_inputs,
        global_sort_indices,
        global_group_sizes,
        global_sorted_experts,
    ) = global_permute_fn(x_TD, selected_experts_TX,
                          moe_instance.num_experts_per_tok,
                          moe_instance.num_local_experts)

    if moe_instance.num_expert_parallelism > 1:
        expert_shard_id = jax.lax.axis_index(moe_instance.expert_axis_name)
        local_expert_size = moe_instance.num_local_experts // moe_instance.num_expert_parallelism

        if moe_instance.is_batch_sharded_by_expert:
            # 2a. Send Tokens To Experts (All-to-All)
            all_shards_group_sizes = jax.lax.all_gather(
                global_group_sizes, axis_name=moe_instance.data_axis_name)

            all_shards_group_sizes_per_expert_shard = jnp.sum(
                all_shards_group_sizes.reshape(
                    moe_instance.num_expert_parallelism,
                    moe_instance.num_expert_parallelism, local_expert_size),
                axis=2)

            input_offsets, send_sizes, output_offsets, recv_sizes = get_all_to_all_params_fn(
                all_shards_group_sizes_per_expert_shard, expert_shard_id,
                moe_instance.num_expert_parallelism)

            local_total_assignments = x_TD.shape[
                0] * moe_instance.num_experts_per_tok
            global_total_assignments = local_total_assignments * moe_instance.num_expert_parallelism
            output_shape_est = jnp.zeros(
                (global_total_assignments, moe_instance.hidden_size),
                dtype=sorted_inputs.dtype)

            inputs_after_all2all = jax.lax.ragged_all_to_all(
                sorted_inputs,
                output_shape_est,
                input_offsets,
                send_sizes,
                output_offsets,
                recv_sizes,
                axis_name=moe_instance.expert_axis_name)

            # 3a. Local Permute
            full_global_group_sizes = jax.lax.all_gather(
                global_group_sizes, axis_name=moe_instance.expert_axis_name)
            (
                compute_inputs,
                local_sorted_indices,
                compute_group_sizes,
                compute_expert_ids,
            ) = local_permute_fn(
                inputs_after_all2all,
                full_global_group_sizes,
                local_expert_size,
                shard_index=expert_shard_id,
                is_offset=False,
            )

        else:
            # 3b. Local "Permute" for Replicated tokens
            (
                compute_inputs,
                local_sorted_indices,
                compute_group_sizes,
                compute_expert_ids,
            ) = local_permute_fn(
                sorted_inputs,
                global_group_sizes[None, :],
                local_expert_size,
                shard_index=expert_shard_id,
                is_offset=True,
                global_sorted_experts=global_sorted_experts,
            )

            reshaped_group_sizes = jnp.sum(global_group_sizes.reshape(
                -1, local_expert_size),
                                           axis=1)
            mask = compute_expert_ids < local_expert_size
            compute_inputs = compute_inputs * mask[..., None]

    else:
        # --- NO EXPERT PARALLELISM ---
        compute_inputs = sorted_inputs
        compute_group_sizes = global_group_sizes
        compute_expert_ids = global_sorted_experts
        local_sorted_indices = jnp.arange(sorted_inputs.shape[0])

    # 4. Compute: Apply experts using Grouped Matrix Multiply
    with jax.named_scope("gating"):
        gating_TEF = gmm_fn(compute_inputs, kernel_gating, compute_group_sizes,
                            moe_instance.tile_size, moe_instance.moe_backend,
                            moe_instance.dtype,
                            moe_instance.qwix_quantized_weight_dtype)
        if moe_instance.edf_sharding[1] is not None:
            gating_TEF = jax.lax.psum(gating_TEF,
                                      axis_name=moe_instance.edf_sharding[1])
        activated_gating_TEF = modeling_flax_utils.ACT2FN[
            moe_instance.hidden_act](gating_TEF)

    with jax.named_scope("up_projection"):
        up_proj_TEF = gmm_fn(compute_inputs, kernel_up_proj,
                             compute_group_sizes, moe_instance.tile_size,
                             moe_instance.moe_backend, moe_instance.dtype,
                             moe_instance.qwix_quantized_weight_dtype)
        if moe_instance.edf_sharding[1] is not None:
            up_proj_TEF = jax.lax.psum(up_proj_TEF,
                                       axis_name=moe_instance.edf_sharding[1])

    fuse_TEF = activated_gating_TEF * up_proj_TEF

    with jax.named_scope("down_projection"):
        intermediate_output = gmm_fn(fuse_TEF, kernel_down_proj,
                                     compute_group_sizes,
                                     moe_instance.tile_size,
                                     moe_instance.moe_backend,
                                     moe_instance.dtype,
                                     moe_instance.qwix_quantized_weight_dtype)
        if moe_instance.efd_sharding[1] is not None:
            intermediate_output = jax.lax.psum(
                intermediate_output, axis_name=moe_instance.efd_sharding[1])

    # 5. Return Results (All-to-All)
    if moe_instance.num_expert_parallelism > 1:
        local_total_assignments = x_TD.shape[
            0] * moe_instance.num_experts_per_tok
        output_shape = jnp.zeros(
            (local_total_assignments, moe_instance.hidden_size),
            dtype=intermediate_output.dtype)

        if moe_instance.is_batch_sharded_by_expert:
            local_output = sort_activations_fn(
                intermediate_output, jnp.argsort(local_sorted_indices))

            input_offsets, send_sizes, output_offsets, recv_sizes = get_all_to_all_params_fn(
                jnp.transpose(all_shards_group_sizes),
                expert_shard_id,
                moe_instance.num_expert_parallelism,
            )
            final_intermediate_output = jax.lax.ragged_all_to_all(
                local_output,
                output_shape,
                input_offsets,
                send_sizes,
                output_offsets,
                recv_sizes,
                axis_name=moe_instance.expert_axis_name)
        else:
            input_offsets, send_sizes, output_offsets, recv_sizes = get_all_to_all_params_fn(
                reshaped_group_sizes,
                expert_shard_id,
                moe_instance.num_expert_parallelism,
                is_batch_sharded=False,
            )
            final_intermediate_output = jax.lax.ragged_all_to_all(
                intermediate_output,
                output_shape,
                input_offsets,
                send_sizes,
                output_offsets,
                recv_sizes,
                axis_name=moe_instance.expert_axis_name)
    else:
        final_intermediate_output = intermediate_output

    # 6. Global Unpermute
    with jax.named_scope("unpermute"):
        output_TD = unpermute_fn(final_intermediate_output,
                                 global_sort_indices, router_weights_TX,
                                 moe_instance.num_experts_per_tok,
                                 moe_instance.dtype)

    return output_TD


def sparse_moe_func(weights: UnfusedMoEWeights, x_TD: jax.Array,
                    gating_output: Tuple[jax.Array, jax.Array],
                    layer: Union[FusedMoE, JaxMoE], mesh: Mesh) -> jax.Array:
    assert isinstance(
        weights, UnfusedMoEWeights), "Expected unfused weights for sparse MoE!"
    weights_TX, indices_TX = gating_output
    if layer.qwix_quantized_weight_dtype:
        gating_up_proj_spec = (PartitionSpec(*layer.edf_sharding),
                               PartitionSpec(*layer.edf_sharding))
        down_proj_spec = (PartitionSpec(*layer.efd_sharding),
                          PartitionSpec(*layer.efd_sharding))
    else:
        gating_up_proj_spec = PartitionSpec(*layer.edf_sharding)
        down_proj_spec = PartitionSpec(*layer.efd_sharding)

    in_specs = (
        PartitionSpec(),  # replicated MoE instance
        PartitionSpec(*layer.activation_ffw_td),  # Sharded x_TD
        PartitionSpec(),  # Replicated router_weights_TX
        PartitionSpec(),  # Replicated selected_experts_TX
        gating_up_proj_spec,  # Sharded gating kernel
        gating_up_proj_spec,  # Sharded up-projection kernel
        down_proj_spec,  # Sharded down-projection kernel
    )
    out_specs = PartitionSpec(*layer.activation_ffw_td)

    mapped_moe_fwd = partial(jax.experimental.shard_map.shard_map,
                             mesh=mesh,
                             in_specs=in_specs,
                             out_specs=out_specs,
                             check_rep=False)(sparse_moe_distributed_fwd)

    # TODO (jacobplatin): this is needed because of issues with Qwix quantizing the `shard_map` in SpraseMatmul
    # Basically, during the abstract pass, we need to manually quantize the weights here for Qwix, but we'll
    # override the actual weight/scale during loading (we just need to make sure Qwix quantizes the weight
    # in the first place).
    kernel_gating_EDF = _process_weight_for_qwix(
        layer.qwix_quantized_weight_dtype,
        "kernel_gating_EDF",
        layer.kernel_gating_EDF,
        channelwise_axes=[0, 2],
        tiled_axes={1: 128})
    kernel_up_proj_EDF = _process_weight_for_qwix(
        layer.qwix_quantized_weight_dtype,
        "kernel_up_proj_EDF",
        layer.kernel_up_proj_EDF,
        channelwise_axes=[0, 2],
        tiled_axes={1: 128})
    kernel_down_proj_EFD = _process_weight_for_qwix(
        layer.qwix_quantized_weight_dtype,
        "kernel_down_proj_EFD",
        layer.kernel_down_proj_EFD,
        channelwise_axes=[0, 2],
        tiled_axes={1: 128})

    weights.w1_weight = kernel_gating_EDF
    weights.w2_weight = kernel_up_proj_EDF
    weights.w3_weight = kernel_down_proj_EFD
    # TODO (jacobplatin): support quantization
    return mapped_moe_fwd(layer, x_TD, weights_TX, indices_TX,
                          weights.w1_weight, weights.w2_weight,
                          weights.w3_weight)


def _process_weight_for_qwix(
        qwix_quantized_weight_dtype: jnp.dtype,
        name: str,
        weight_param: Union[ptq.WithAux, nnx.Param],
        channelwise_axes: Optional[List[int]] = [],
        tiled_axes: dict = {}
) -> Union[jax.Array, Tuple[nnx.Param, nnx.Param]]:
    """
    Extracts weight value, applies quantization if needed,
    and returns the underlying array.

    Args:
        qwix_quantized_weight_dtype: The dtype to quantize the weights to.
        name: The name of the weight.
        weight_param: The weight parameter.
        channelwise_axes: The axes to quantize.
        tiled_axes: The axes to tile.

    Returns:
        The quantized weight.
    """

    weight = weight_param.value

    if qwix_quantized_weight_dtype:
        if not isinstance(weight, ptq.WithAux):
            weight = manually_quantize_qwix_weight(
                name, weight, qwix_quantized_weight_dtype, channelwise_axes,
                tiled_axes, "absmax")
        return (weight.array.qvalue, weight.array.scale)

    return weight
